"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.lib.constants import OBJECT_TYPE, TUPLE_TYPE
from nagini_translation.lib.program_nodes import (
    GenericType,
    OptionalType,
    PythonClass,
    PythonType,
    TypeVar,
    UnionType,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.abstract import Context, Expr
from typing import List, Tuple


class TypeDomainFactory:
    """
    Creates domain functions and axioms that represent the Python/mypy type
    system within Viper.
    """
    UNION_TYPE_SIZE = 4

    def __init__(self, viper: ViperAST, translator: 'Translator') -> None:
        self.viper = viper
        self.type_domain = 'PyType'
        self.translator = translator

    def no_position(self, ctx: Context) -> 'silver.ast.Position':
        return self.translator.no_position(ctx)

    def to_position(self, node: ast.AST, ctx: Context) -> 'silver.ast.Position':
        return self.translator.to_position(node, ctx)

    def no_info(self, ctx: Context) -> 'silver.ast.Position':
        return self.translator.no_info(ctx)

    def get_default_axioms(self,
                           ctx: Context) -> List['silver.ast.DomainAxiom']:
        result = [
            self.create_transitivity_axiom(ctx),
            self.create_reflexivity_axiom(ctx),
            self.create_extends_implies_subtype_axiom(ctx),
            self.create_null_type_axiom(ctx),
            self.create_object_subtype_axiom(ctx),
            self.create_subtype_exclusion_axiom(ctx),
            self.create_subtype_exclusion_axiom_2(ctx),
            self.create_subtype_exclusion_propagation_axiom(ctx),
            self.create_tuple_arg_axiom(ctx),
            self.create_tuple_args_axiom(ctx),
            self.create_tuple_subtype_axiom(ctx),
        ]
        result.extend(self.create_union_subtype_axioms(ctx))
        result.extend(self.create_subtype_union_axioms(ctx))
        return result

    def get_default_functions(self,
                              ctx: Context) -> List['silver.ast.DomainFunc']:
        result = [
            self.extends_func(ctx),
            self.issubtype_func(ctx),
            self.isnotsubtype_func(ctx),
            self.tuple_args_func(ctx),
            self.typeof_func(ctx),
            self.basic_func(ctx),
        ]
        result.extend(self.union_funcs(ctx))
        return result

    def union_funcs(self, ctx: Context) -> List['silver.ast.DomainFunc']:
        """
        Creates UNION_TYPE_SIZE functions of the following form:
        function union_type_n(arg_1: PyType, ..., arg_n: PyType): PyType
        """
        result = []
        position, info = self.no_position(ctx), self.no_info(ctx)
        args = []
        for i in range(1, self.UNION_TYPE_SIZE + 1):
            arg = self.viper.LocalVarDecl('arg_' + str(i), self.type_type(),
                                          position, info)
            args.append(arg)
            func = self.viper.DomainFunc('union_type_' + str(i), args,
                                         self.type_type(), False, position,
                                         info, self.type_domain)
            result.append(func)
        return result

    def create_union_subtype_axioms(self,
            ctx: Context) -> List['silver.ast.DomainAxiom']:
        """
        Creates UNION_TYPE_SIZE axioms of the following form:
        (forall arg_1: PyType, ..., arg_n: PyType, X: PyType ::
        { issubtype(X, union_type_n(arg_1, ..., arg_n)) }
          issubtype(X, union_type_n(arg_1, ..., arg_n)) ==
          (false || issubtype(X, arg_1) || ... || issubtype(X, arg_n)))
        """
        result = []
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_decls = []
        args = []
        arg_subtype = self.viper.FalseLit(position, info)
        sub_decl = self.viper.LocalVarDecl('X', self.type_type(), position,
                                           info)
        sub_var = self.viper.LocalVar('X', self.type_type(), position, info)
        for i in range(1, self.UNION_TYPE_SIZE + 1):
            arg_decl = self.viper.LocalVarDecl('arg_' + str(i), self.type_type(),
                                                position, info)
            arg = self.viper.LocalVar('arg_' + str(i), self.type_type(), position,
                                      info)
            arg_decls.append(arg_decl)
            args.append(arg)
            union = self.viper.DomainFuncApp('union_type_' + str(i), args,
                                             self.type_type(), position,
                                             info, self.type_domain)
            subtype_union = self._issubtype(sub_var, union, ctx, position)
            current_arg_subtype = self._issubtype(sub_var, arg, ctx, position)
            arg_subtype = self.viper.Or(arg_subtype, current_arg_subtype,
                                        position, info)
            body = self.viper.EqCmp(subtype_union, arg_subtype, position, info)
            trigger = self.viper.Trigger([subtype_union], position, info)
            forall = self.viper.Forall(arg_decls + [sub_decl], [trigger], body,
                                       position, info)
            axiom = self.viper.DomainAxiom('union_subtype_' + str(i), forall,
                                           position, info, self.type_domain)
            result.append(axiom)
        return result

    def create_subtype_union_axioms(self,
                                    ctx: Context) -> List['silver.ast.DomainAxiom']:
        """
            Creates UNION_TYPE_SIZE axioms of the following form:
            (forall arg_1: PyType, ..., arg_n: PyType, X: PyType ::
            { issubtype(union_type_n(arg_1, ..., arg_n), X) }
              issubtype(union_type_n(arg_1, ..., arg_n), X) ==
              (true && issubtype(arg_1, X) && ... && issubtype(arg_n, X)))
            """
        result = []
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_decls = []
        args = []
        arg_subtype = self.viper.TrueLit(position, info)
        sub_decl = self.viper.LocalVarDecl('X', self.type_type(), position,
                                           info)
        sub_var = self.viper.LocalVar('X', self.type_type(), position, info)
        for i in range(1, self.UNION_TYPE_SIZE + 1):
            arg_decl = self.viper.LocalVarDecl('arg_' + str(i), self.type_type(),
                                               position, info)
            arg = self.viper.LocalVar('arg_' + str(i), self.type_type(), position,
                                      info)
            arg_decls.append(arg_decl)
            args.append(arg)
            union = self.viper.DomainFuncApp('union_type_' + str(i), args,
                                             self.type_type(), position,
                                             info, self.type_domain)
            subtype_union = self._issubtype(union, sub_var, ctx, position)
            current_arg_subtype = self._issubtype(arg, sub_var, ctx, position)
            arg_subtype = self.viper.And(arg_subtype, current_arg_subtype,
                                         position, info)
            body = self.viper.EqCmp(subtype_union, arg_subtype, position, info)
            trigger = self.viper.Trigger([subtype_union], position, info)
            forall = self.viper.Forall(arg_decls + [sub_decl], [trigger], body,
                                       position, info)
            axiom = self.viper.DomainAxiom('subtype_union_' + str(i), forall,
                                           position, info, self.type_domain)
            result.append(axiom)
        return result

    def create_tuple_subtype_axiom(self,
                                   ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        (forall seq1: Seq[PyType], seq2: Seq[PyType] ::
            seq1 != seq2 && |seq1| == |seq2| &&
            (forall i: Int :: i >= 0 && i < |seq1| ==>
                              issubtype(seq1[i], seq2[i]))
            ==> issubtype(tuple(seq1), tuple(seq2)))

        """
        name = 'tuple_self_subtype'
        position, info = self.no_position(ctx), self.no_info(ctx)
        seq_type = self.viper.SeqType(self.type_type())
        seq1_decl = self.viper.LocalVarDecl('seq1', seq_type, position, info)
        seq1_ref = self.viper.LocalVar('seq1', seq_type, position, info)
        seq2_decl = self.viper.LocalVarDecl('seq2', seq_type, position, info)
        seq2_ref = self.viper.LocalVar('seq2', seq_type, position, info)
        tuple1 = self.viper.DomainFuncApp('tuple', [seq1_ref],
                                          self.type_type(), position,
                                          info, self.type_domain)
        tuple2 = self.viper.DomainFuncApp('tuple', [seq2_ref],
                                          self.type_type(), position,
                                          info, self.type_domain)

        subtype = self._issubtype(tuple1, tuple2, ctx)
        length1 = self.viper.SeqLength(seq1_ref, position, info)
        length2 = self.viper.SeqLength(seq2_ref, position, info)
        length_eq = self.viper.EqCmp(length1, length2, position, info)
        i_decl = self.viper.LocalVarDecl('i', self.viper.Int, position, info)
        i_ref = self.viper.LocalVar('i', self.viper.Int, position, info)
        i1 = self.viper.SeqIndex(seq1_ref, i_ref, position, info)
        i2 = self.viper.SeqIndex(seq2_ref, i_ref, position, info)
        i_subtype = self._issubtype(i1, i2, ctx)
        i_nonneg = self.viper.GeCmp(i_ref, self.viper.IntLit(0, position, info),
                                    position, info)
        i_lt_length = self.viper.LtCmp(i_ref, length1, position, info)
        i_in_bounds = self.viper.And(i_nonneg, i_lt_length, position, info)
        forall_body = self.viper.Implies(i_in_bounds, i_subtype, position, info)
        rhs_forall = self.viper.Forall([i_decl], [], forall_body, position,
                                       info)
        seqs_different = self.viper.NeCmp(seq1_ref, seq2_ref, position, info)
        seq_restrictions = self.viper.And(seqs_different, length_eq, position,
                                          info)
        rhs = self.viper.And(seq_restrictions, rhs_forall, position, info)
        body = self.viper.Implies(rhs, subtype, position, info)
        body = self.viper.Forall([seq1_decl, seq2_decl], [], body, position,
                                 info)
        result = self.viper.DomainAxiom(name, body, position, info,
                                        self.type_domain)
        return result

    def create_tuple_arg_axiom(self,
                               ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        (forall seq: Seq[PyType], i: Int, Z: PyType ::
          { tuple(seq),tuple_arg(Z, i) }
          issubtype(Z, tuple(seq)) ==> issubtype(tuple_arg(Z, i), seq[i]))
        """
        name = 'tuple_arg_def'
        position, info = self.no_position(ctx), self.no_info(ctx)
        seq_type = self.viper.SeqType(self.type_type())
        z_decl = self.viper.LocalVarDecl('Z', self.type_type(), position, info)
        z_ref = self.viper.LocalVar('Z', self.type_type(), position, info)
        seq_decl = self.viper.LocalVarDecl('seq', seq_type, position, info)
        seq_ref = self.viper.LocalVar('seq', seq_type, position, info)
        i_decl = self.viper.LocalVarDecl('i', self.viper.Int, position, info)
        i_ref = self.viper.LocalVar('i', self.viper.Int, position, info)
        tuple_seq = self.viper.DomainFuncApp('tuple', [seq_ref],
                                             self.type_type(), position, info,
                                             self.type_domain)
        subtype = self._issubtype(z_ref, tuple_seq, ctx)
        arg_args = [z_ref, i_ref]
        arg_func = self.viper.DomainFuncApp('tuple_arg', arg_args,
                                             self.type_type(), position, info,
                                            self.type_domain)
        seq_index = self.viper.SeqIndex(seq_ref, i_ref, position, info)
        rhs = self._issubtype(arg_func, seq_index, ctx)

        body = self.viper.Implies(subtype, rhs, position, info)
        trigger = self.viper.Trigger([tuple_seq, arg_func], position, info)
        body = self.viper.Forall([seq_decl, i_decl, z_decl], [trigger], body,
                                 position, info)
        result = self.viper.DomainAxiom(name, body, position, info,
                                        self.type_domain)
        return result

    def create_tuple_args_axiom(self,
                                   ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        (forall seq: Seq[PyType], Z: PyType :: { issubtype(Z, tuple(seq)) }
          issubtype(Z, tuple(seq)) ==> |tuple_args(Z)| == |seq|)
        """
        name = 'tuple_args_def'
        position, info = self.no_position(ctx), self.no_info(ctx)
        seq_type = self.viper.SeqType(self.type_type())
        z_decl = self.viper.LocalVarDecl('Z', self.type_type(), position, info)
        z_ref = self.viper.LocalVar('Z', self.type_type(), position, info)
        seq_decl = self.viper.LocalVarDecl('seq', seq_type, position, info)
        seq_ref = self.viper.LocalVar('seq', seq_type, position, info)
        tuple_seq = self.viper.DomainFuncApp('tuple', [seq_ref],
                                             self.type_type(), position, info,
                                             self.type_domain)
        subtype = self._issubtype(z_ref, tuple_seq, ctx)
        args_func = self.viper.DomainFuncApp('tuple_args', [z_ref],
                                             seq_type, position, info,
                                             self.type_domain)
        args_func_len = self.viper.SeqLength(args_func, position, info)
        seq_ref_len = self.viper.SeqLength(seq_ref, position, info)
        rhs = self.viper.EqCmp(args_func_len, seq_ref_len, position, info)
        body = self.viper.Implies(subtype, rhs, position, info)
        trigger = self.viper.Trigger([subtype], position, info)
        body = self.viper.Forall([seq_decl, z_decl], [trigger], body, position,
                                 info)
        result = self.viper.DomainAxiom(name, body, position, info,
                                        self.type_domain)
        return result

    def tuple_args_func(self, ctx: Context) -> 'silver.ast.DomainFunc':
        position, info = self.no_position(ctx), self.no_info(ctx)
        args = [self.viper.LocalVarDecl('t', self.type_type(), position, info)]
        result = self.viper.DomainFunc('tuple_args', args,
                                       self.viper.SeqType(self.type_type()),
                                       False, position, info, self.type_domain)
        return result

    def create_type(self, cls: 'PythonClass',
                    ctx: Context) -> Tuple['silver.ast.DomainFunc',
                                           'silver.ast.DomainAxiom']:
        """
        Creates the type domain function and subtype axiom for this class
        """
        if not cls.superclass:
            supertype = cls.module.global_module.classes[OBJECT_TYPE]
        else:
            supertype = cls.superclass
        position = self.to_position(cls.node, ctx)
        info = self.no_info(ctx)
        type_nargs = len(cls.type_vars) if cls.name != TUPLE_TYPE else -1
        type_funcs = self.create_type_function(cls.sil_name, type_nargs,
                                               position, info, ctx)
        if cls.interface and not cls.superclass:
            subtype_axiom = None
        else:
            subtype_axiom = self.create_subtype_axiom(cls, supertype,
                                                      position, info, ctx)
        funcs = type_funcs
        axioms = [subtype_axiom] if subtype_axiom else []
        if cls.type_vars or cls.name == TUPLE_TYPE:
            funcs.extend(self.create_arg_functions(cls, ctx))
            # create functions for type arguments
            axioms.extend(self.create_arg_axioms(cls, ctx))
            # create axioms that relate arguments to functions
        return funcs, axioms

    def create_arg_functions(self, cls: 'PythonClass',
                             ctx: Context) -> List['silver.ast.DomainFunc']:
        position, info = self.no_position(ctx), self.no_info(ctx)
        result = []
        name = cls.sil_name + '_arg'
        args = []
        args.append(
            self.viper.LocalVarDecl('typ', self.type_type(), position, info))
        args.append(
            self.viper.LocalVarDecl('index', self.viper.Int, position, info))
        result = self.viper.DomainFunc(name, args, self.type_type(), False,
                                       position, info, self.type_domain)
        return [result]

    def create_arg_axioms(self, cls: 'PythonClass',
                          ctx: Context) -> List['silver.ast.DomainAxiom']:
        """
        Creates an axiom defining the type argument getter functions for a given
        type, e.g. for PSeq:
        (forall Z: PyType, arg0: PyType ::
          issubtype(Z, PSeq(arg0)) ==> PSeq_arg(Z, 0) == arg0)
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        z_decl = self.viper.LocalVarDecl('Z', self.type_type(), position, info)
        z_ref = self.viper.LocalVar('Z', self.type_type(), position, info)
        decls = [z_decl]
        type_args = []
        for i, var in enumerate(cls.type_vars):
            name = 'arg' + str(i)
            decl = self.viper.LocalVarDecl(name, self.type_type(), position,
                                           info)
            ref = self.viper.LocalVar(name, self.type_type(), position, info)
            decls.append(decl)
            type_args.append(ref)
        result = []
        func_lit = self.viper.DomainFuncApp(cls.sil_name, type_args,
                                            self.type_type(), position, info,
                                            self.type_domain)
        subtype = self._issubtype(z_ref, func_lit, ctx)
        for i, var in enumerate(cls.type_vars):
            args = [z_ref, self.viper.IntLit(i, position, info)]
            current_arg = self.viper.DomainFuncApp(cls.sil_name + '_arg', args,
                                                   self.type_type(), position,
                                                   info, self.type_domain)
            if cls.name == TUPLE_TYPE:
                args = [current_arg, type_args[i]]
                rhs = self._issubtype(current_arg, type_args[i], ctx)
            else:
                rhs = self.viper.EqCmp(current_arg, type_args[i], position,
                                       info)
            implication = self.viper.Implies(subtype, rhs, position, info)
            quantifier = self.viper.Forall(decls, [], implication, position,
                                           info)
            axiom = self.viper.DomainAxiom(cls.sil_name + '_args' + str(i),
                                           quantifier, position, info,
                                           self.type_domain)
            result.append(axiom)
        return result

    def create_type_function(self, name: str, type_nargs: int,
                             position: 'silver.ast.Position',
                             info: 'silver.ast.Info',
                             ctx: Context) -> List['silver.ast.DomainFunc']:
        args = []
        result = []
        if type_nargs != 0:
            result.append(self.viper.DomainFunc(name + '_basic', [], self.type_type(),
                                                True, position, info, self.type_domain))
        if type_nargs == -1:
            seq_type = self.viper.SeqType(self.type_type())
            args.append(self.viper.LocalVarDecl('args', seq_type, position,
                                                info))
        else:
            for i in range(type_nargs):
                args.append(self.viper.LocalVarDecl('arg' + str(i),
                                                    self.type_type(), position,
                                                    info))
        result.append(self.viper.DomainFunc(name, args, self.type_type(),
                                            len(args) == 0, position, info,
                                            self.type_domain))
        return result

    def create_type_domain(self, type_funcs: List['silver.ast.DomainFunc'],
                           type_axioms: List['silver.ast.DomainAxiom'],
                           ctx: Context) -> 'silver.ast.Domain':
        result = self.viper.Domain(self.type_domain, type_funcs, type_axioms,
                                   [], self.no_position(ctx), self.no_info(ctx))
        return result

    def type_type(self) -> 'silver.ast.DomainType':
        """
        Creates a reference to the domain type we use for the Python types
        """
        return self.viper.DomainType(self.type_domain, {}, [])

    def _issubtype(self, sub: Expr, super: Expr, ctx: Context,
                   position=None) -> 'silver.ast.DomainFuncApp':
        return self._issubtype_app('issubtype', sub, super, ctx, position)

    def _extends(self, sub: Expr, super: Expr, ctx: Context,
                 position=None) -> 'silver.ast.DomainFuncApp':
        return self._issubtype_app('extends_', sub, super, ctx, position)

    def _isnotsubtype(self, sub: Expr, super: Expr, ctx: Context,
                      position=None) -> 'silver.ast.DomainFuncApp':
        return self._issubtype_app('isnotsubtype', sub, super, ctx, position)

    def _issubtype_app(self, name: str, sub: Expr, super: Expr, ctx: Context,
                       position=None) -> 'silver.ast.DomainFuncApp':
        if not position:
            position = self.no_position(ctx)
        return self.viper.DomainFuncApp(name, [sub, super],
                                        self.viper.Bool, position,
                                        self.no_info(ctx), self.type_domain)

    def typeof(self, arg: Expr, ctx: Context) -> 'silver.ast.DomainFuncApp':
        return self.viper.DomainFuncApp('typeof', [arg], self.type_type(),
                                        self.no_position(ctx),
                                        self.no_info(ctx),
                                        self.type_domain)

    def create_subtype_axiom(self, type: PythonType, supertype: PythonType,
                             position: 'silver.ast.Position',
                             info: 'silver.ast.Info',
                             ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates a domain axiom that indicates a subtype relationship
        between type and supertype, quantified over type variables.

        E.g. for class Sub(Generic[T], Super[T, int]):
        forall arg0: PyType :: {Sub(arg0)}
          extends_(Sub(arg0), Super(arg0, int()))
        """
        type_arg_decls = []
        type_args = []
        ctx.bound_type_vars = {}
        if type.name == TUPLE_TYPE:
            seq_type = self.viper.SeqType(self.type_type())
            args_ref = self.viper.LocalVar('args', seq_type, position, info)
            type_args.append(args_ref)
            type_arg_decls.append(
                self.viper.LocalVarDecl('args', seq_type, position, info))
        else:
            for name, var in type.type_vars.items():
                var = self.viper.LocalVar(name, self.type_type(), position,
                                          info)
                type_args.append(var)
                type_arg_decls.append(self.viper.LocalVarDecl(name,
                                                              self.type_type(),
                                                              position, info))
                ctx.bound_type_vars[(type.name, name)] = var

        type_var = self.viper.LocalVar('class', self.type_type(), position,
                                       info)
        type_func = self.viper.DomainFuncApp(type.sil_name, type_args,
                                             self.type_type(), position, info,
                                             self.type_domain)

        supertype_func = self.translate_type_literal(supertype,
                                                     position, ctx)
        body = self._extends(type_func, supertype_func, ctx, position)
        if type.name == TUPLE_TYPE:
            # Special case for tuples:
            # forall args: Seq[PyType] :: { tuple(args) }
            # (forall e: PyType :: (e in args) ==> e == object())
            #                       ==> extends_(tuple(args), object())
            # This way, tuple(Seq(object(), object())) extends object,
            # but tuple(int(), int()) does not, which we need, because types
            # that extend the same supertype are said not to be subtypes of
            # each other by issubtype_exclusion.
            e_decl = self.viper.LocalVarDecl('e', self.type_type(), position,
                                             info)
            e_ref = self.viper.LocalVar('e', self.type_type(), position, info)
            e_contained = self.viper.SeqContains(e_ref, args_ref, position,
                                                 info)
            object_func = self.viper.DomainFuncApp('object', [],
                                                   self.type_type(), position,
                                                   info, self.type_domain)
            e_is_object = self.viper.EqCmp(e_ref, object_func, position, info)
            implication = self.viper.Implies(e_contained, e_is_object, position,
                                             info)
            body_lhs = self.viper.Forall([e_decl], [], implication, position,
                                         info)
            body = self.viper.Implies(body_lhs, body, position, info)
        if type_arg_decls:
            basic_type = self.viper.DomainFuncApp(type.sil_name + '_basic', [],
                                                  self.type_type(), position, info,
                                                  self.type_domain)
        else:
            basic_type = type_func
        type_func_basic = self.viper.DomainFuncApp('get_basic', [type_func],
                                                   self.type_type(), position, info,
                                                   self.type_domain)
        define_basic = self.viper.EqCmp(type_func_basic, basic_type, position, info)
        body = self.viper.And(body, define_basic, position, info)
        if type_arg_decls:
            trigger = self.viper.Trigger([type_func], position, info)
            body = self.viper.Forall(type_arg_decls, [trigger], body, position,
                                     info)
        return self.viper.DomainAxiom('subtype_' + type.sil_name, body,
                                      position, info,
                                      self.type_domain)

    def create_extends_implies_subtype_axiom(self,
            ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates an axiom that states that an extends-relationship between two
        types implies a subtype-relationship:

        forall sub: PyType, sub2: PyType :: { extends_(sub, sub2) }
        extends_(sub, sub2)
        ==>
        issubtype(sub, sub2)
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_sub = self.viper.LocalVarDecl('sub', self.type_type(), position,
                                          info)
        var_sub = self.viper.LocalVar('sub', self.type_type(), position, info)
        arg_super = self.viper.LocalVarDecl('sub2', self.type_type(), position,
                                            info)
        var_super = self.viper.LocalVar('sub2', self.type_type(), position,
                                        info)
        extends = self._extends(var_sub, var_super, ctx)
        subtype = self._issubtype(var_sub, var_super, ctx)
        implication = self.viper.Implies(extends, subtype, position, info)
        trigger = self.viper.Trigger([extends], position, info)
        body = self.viper.Forall([arg_sub, arg_super], [trigger],
                                 implication, position, info)
        return self.viper.DomainAxiom('extends_implies_subtype', body,
                                      position, info, self.type_domain)

    def create_subtype_exclusion_axiom_2(self,
            ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        forall sub: PyType, super: PyType ::
        { issubtype(sub, super) } { issubtype(super, sub) }
        issubtype(sub, super) && sub != super ==> !issubtype(super, sub)
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_sub = self.viper.LocalVarDecl('sub', self.type_type(),
                                          position, info)
        var_sub = self.viper.LocalVar('sub', self.type_type(), position, info)
        arg_super = self.viper.LocalVarDecl('super', self.type_type(),
                                            position, info)
        var_super = self.viper.LocalVar('super', self.type_type(),
                                        position, info)
        sub_super = self._issubtype(var_sub, var_super, ctx)
        super_sub = self._issubtype(var_super, var_sub, ctx)
        not_super_sub = self.viper.Not(super_sub, position, info)
        not_equal = self.viper.NeCmp(var_sub, var_super, position, info)
        lhs = self.viper.And(sub_super, not_equal, position, info)
        implication = self.viper.Implies(lhs, not_super_sub, position, info)
        trigger = self.viper.Trigger([sub_super], position, info)
        trigger2 = self.viper.Trigger([super_sub], position, info)
        body = self.viper.Forall([arg_sub, arg_super], [trigger, trigger2],
                                 implication, position, info)
        return self.viper.DomainAxiom('issubtype_exclusion_2', body,
                                      position, info, self.type_domain)

    def create_subtype_exclusion_axiom(
            self, ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates an axiom that states that two types that directly extend
        another type cannot be subtypes of each other:

        forall sub: PyType, sub2: PyType, super: PyType ::
        { extends_(sub, super),extends_(sub2, super) }
        extends_(sub, super) && extends_(sub2, super) && (sub != sub2)
        ==>
        isnotsubtype(sub, sub2) && isnotsubtype(sub2, sub))
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_sub = self.viper.LocalVarDecl('sub', self.type_type(), position,
                                          info)
        var_sub = self.viper.LocalVar('sub', self.type_type(),
                                      position, info)
        arg_sub2 = self.viper.LocalVarDecl('sub2', self.type_type(),
                                           position, info)
        var_sub2 = self.viper.LocalVar('sub2', self.type_type(), position, info)
        arg_super = self.viper.LocalVarDecl('super', self.type_type(),
                                            position, info)
        var_super = self.viper.LocalVar('super', self.type_type(), position,
                                        info)

        sub_super = self._extends(var_sub, var_super, ctx)
        sub2_super = self._extends(var_sub2, var_super, ctx)
        sub_sub2 = self._isnotsubtype(var_sub, var_sub2, ctx)
        sub2_sub = self._isnotsubtype(var_sub2, var_sub, ctx)
        not_subtypes = self.viper.And(sub_sub2, sub2_sub, position, info)
        subs_not_equal = self.viper.NeCmp(var_sub, var_sub2, position, info)
        extends = self.viper.And(sub_super, sub2_super, position, info)
        lhs = self.viper.And(extends, subs_not_equal, position, info)
        implication = self.viper.Implies(lhs, not_subtypes, position, info)
        trigger = self.viper.Trigger([sub_super, sub2_super], position, info)
        body = self.viper.Forall([arg_sub, arg_sub2, arg_super], [trigger],
                                 implication, position, info)
        return self.viper.DomainAxiom('issubtype_exclusion', body,
                                      position, info, self.type_domain)

    def create_subtype_exclusion_propagation_axiom(self,
            ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates an axiom that propagates the information that two types
        are not subtypes down the type hierarchy:

        forall sub: PyType, middle: PyType, super: PyType ::
        { issubtype(sub, middle),isnotsubtype(middle, super) }
        issubtype(sub, middle) && isnotsubtype(middle, super)
        ==>
        !issubtype(sub, super))
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_sub = self.viper.LocalVarDecl('sub', self.type_type(), position,
                                          info)
        var_sub = self.viper.LocalVar('sub', self.type_type(), position, info)
        arg_middle = self.viper.LocalVarDecl('middle', self.type_type(),
                                             position, info)
        var_middle = self.viper.LocalVar('middle', self.type_type(),
                                         position, info)
        arg_super = self.viper.LocalVarDecl('super', self.type_type(),
                                            position, info)
        var_super = self.viper.LocalVar('super', self.type_type(),
                                        position, info)

        sub_middle = self._issubtype(var_sub, var_middle, ctx)

        middle_super = self._isnotsubtype(var_middle, var_super, ctx)
        sub_super = self._issubtype(var_sub, var_super, ctx)

        not_sub_super = self.viper.Not(sub_super, position, info)
        implication = self.viper.Implies(
            self.viper.And(sub_middle, middle_super, position, info),
            not_sub_super, position, info)
        trigger = self.viper.Trigger([sub_middle, middle_super],
                                     position, info)
        body = self.viper.Forall([arg_sub, arg_middle, arg_super], [trigger],
                                 implication, position, info)
        return self.viper.DomainAxiom('issubtype_exclusion_propagation', body,
                                      position, info, self.type_domain)

    def create_null_type_axiom(self, ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates an axiom that states that the type of null is None:

        forall r: Ref :: { typeof(r) }
          issubtype(typeof(r), NoneType()) == (r == null)
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_r = self.viper.LocalVarDecl('r', self.viper.Ref, position, info)
        var_r = self.viper.LocalVar('r', self.viper.Ref, position, info)
        none_type = self.viper.DomainFuncApp('NoneType', [], self.type_type(),
                                             position, info, self.type_domain)
        typeof = self.typeof(var_r, ctx)
        subtype = self._issubtype(typeof, none_type, ctx)
        is_null = self.viper.EqCmp(var_r,
                                   self.viper.NullLit(position, info),
                                   position, info)
        biimplication = self.viper.EqCmp(subtype, is_null, position, info)
        trigger = self.viper.Trigger([typeof], position, info)
        body = self.viper.Forall([arg_r], [trigger],
                                 biimplication, position, info)
        return self.viper.DomainAxiom('null_nonetype', body, position, info,
                                      self.type_domain)

    def create_object_type(self, ctx: Context) -> 'silver.ast.DomainFunc':
        return self.create_type_function(OBJECT_TYPE, 0, self.no_position(ctx),
                                         self.no_info(ctx))[0]

    def create_null_type(self, ctx: Context) -> 'silver.ast.DomainFunc':
        return self.create_type_function('NoneType', 0, self.no_position(ctx),
                                         self.no_info(ctx))[0]

    def create_transitivity_axiom(self,
                                  ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates the transitivity axiom for the PyType domain:
        forall sub: PyType, middle: PyType, super: PyType ::
            { issubtype(sub, middle),issubtype(middle, super) }
            issubtype(sub, middle) && issubtype(middle, super)
            ==>
            issubtype(sub, super)
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg_sub = self.viper.LocalVarDecl('sub', self.type_type(),
                                          position, info)
        var_sub = self.viper.LocalVar('sub', self.type_type(), position, info)
        arg_middle = self.viper.LocalVarDecl('middle', self.type_type(),
                                             position, info)
        var_middle = self.viper.LocalVar('middle', self.type_type(),
                                         position, info)
        arg_super = self.viper.LocalVarDecl('super', self.type_type(),
                                            position, info)
        var_super = self.viper.LocalVar('super', self.type_type(),
                                        position, info)

        sub_middle = self._issubtype(var_sub, var_middle, ctx)

        middle_super = self._issubtype(var_middle, var_super, ctx)

        sub_super = self._issubtype(var_sub, var_super, ctx)

        implication = self.viper.Implies(
            self.viper.And(sub_middle, middle_super, position, info),
            sub_super, position, info)
        trigger = self.viper.Trigger([sub_middle, middle_super], position, info)
        body = self.viper.Forall([arg_sub, arg_middle, arg_super], [trigger],
                                 implication, position, info)
        return self.viper.DomainAxiom('issubtype_transitivity', body,
                                      position, info, self.type_domain)

    def create_reflexivity_axiom(self,
                                 ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates the reflexivity axiom for the PyType domain:
        forall type: PyType :: { issubtype(type, type) } issubtype(type, type)
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg = self.viper.LocalVarDecl('type_', self.type_type(), position, info)
        var = self.viper.LocalVar('type_', self.type_type(), position, info)
        reflexive_subtype = self._issubtype(var, var, ctx)
        trigger_exp = reflexive_subtype
        trigger = self.viper.Trigger([trigger_exp], position, info)
        body = self.viper.Forall([arg], [trigger], reflexive_subtype,
                                 position, info)
        return self.viper.DomainAxiom('issubtype_reflexivity', body,
                                      position, info, self.type_domain)

    def create_object_subtype_axiom(self,
                                    ctx: Context) -> 'silver.ast.DomainAxiom':
        """
        Creates the axiom saying that all types are subtypes of object:
        forall type: PyType :: { issubtype(type, object()) }
        issubtype(type, object())
        """
        position, info = self.no_position(ctx), self.no_info(ctx)
        arg = self.viper.LocalVarDecl('type_', self.type_type(), position, info)
        var = self.viper.LocalVar('type_', self.type_type(), position, info)
        object_type = self.viper.DomainFuncApp('object', [], self.type_type(),
                                               position, info, self.type_domain)
        object_subtype = self._issubtype(var, object_type, ctx)
        trigger_exp = object_subtype
        trigger = self.viper.Trigger([trigger_exp], position, info)
        body = self.viper.Forall([arg], [trigger], object_subtype, position,
                                 info)
        return self.viper.DomainAxiom('issubtype_object', body, position, info,
                                      self.type_domain)

    def typeof_func(self, ctx: Context) -> 'silver.ast.DomainFunc':
        """
        Creates the typeof domain function
        """
        obj_var = self.viper.LocalVarDecl('obj', self.viper.Ref,
                                          self.no_position(ctx),
                                          self.no_info(ctx))
        return self.viper.DomainFunc('typeof', [obj_var],
                                     self.type_type(), False,
                                     self.no_position(ctx), self.no_info(ctx),
                                     self.type_domain)

    def basic_func(self, ctx: Context) -> 'silver.ast.DomainFunc':
        """
        Creates the get_basic domain function
        """
        t_var = self.viper.LocalVarDecl('t', self.type_type(),
                                          self.no_position(ctx),
                                          self.no_info(ctx))
        return self.viper.DomainFunc('get_basic', [t_var],
                                     self.type_type(), False,
                                     self.no_position(ctx), self.no_info(ctx),
                                     self.type_domain)

    def subtype_func(self, name: str, ctx: Context) -> 'silver.ast.DomainFunc':
        """
        Creates the issubtype, extends and isnotsubtype domain functions.
        """
        sub_var = self.viper.LocalVarDecl('sub', self.type_type(),
                                          self.no_position(ctx),
                                          self.no_info(ctx))
        super_var = self.viper.LocalVarDecl('super', self.type_type(),
                                            self.no_position(ctx),
                                            self.no_info(ctx))
        return self.viper.DomainFunc(name, [sub_var, super_var],
                                     self.viper.Bool, False,
                                     self.no_position(ctx), self.no_info(ctx),
                                     self.type_domain)

    def issubtype_func(self, ctx: Context) -> 'silver.ast.DomainFunc':
        return self.subtype_func('issubtype', ctx)

    def isnotsubtype_func(self, ctx: Context) -> 'silver.ast.DomainFunc':
        return self.subtype_func('isnotsubtype', ctx)

    def extends_func(self, ctx: Context) -> 'silver.ast.DomainFunc':
        return self.subtype_func('extends_', ctx)

    def dynamic_type_check(self, lhs: 'Expr',
                           type: 'Expr', position: 'silver.ast.Position',
                           ctx: Context):
        type_func = self.typeof(lhs, ctx)
        result = self.viper.EqCmp(type_func, type, self.no_position(ctx),
                                  self.no_info(ctx))
        return result

    def subtype_check(self, type_func: 'Expr', type: 'PythonType',
                      position: 'silver.ast.Position',
                      ctx: Context, concrete: bool = False) -> Expr:
        """
        Creates an expression which denotes if the given ``type_func``
        expression is a subtype of (or, if ``concrete`` is true, equivalent to)
        the given ``type``.
        """
        info = self.no_info(ctx)
        if isinstance(type, GenericType) and not type.exact_length:
            assert type.name == TUPLE_TYPE
            seq_type = self.viper.SeqType(self.type_type())
            i_ref = self.viper.LocalVar('i', self.viper.Int, position, info)
            i_decl = self.viper.LocalVarDecl('i', self.viper.Int, position,
                                             info)
            zero = self.viper.IntLit(0, position, info)
            i_ge_zero = self.viper.GeCmp(i_ref, zero, position, info)
            tuple_args = self.viper.DomainFuncApp('tuple_args', [type_func],
                                                  seq_type, position, info,
                                                  self.type_domain)
            tuple_args_len = self.viper.SeqLength(tuple_args, position, info)
            i_lt_len = self.viper.LtCmp(i_ref, tuple_args_len, position, info)
            i_in_bounds = self.viper.And(i_ge_zero, i_lt_len, position, info)
            tuple_arg = self.viper.DomainFuncApp('tuple_arg',
                                                 [type_func, i_ref],
                                                 self.type_type(), position,
                                                 info, self.type_domain)
            arg_lit = self.translate_type_literal(type.type_args[0], position,
                                                  ctx)
            if concrete:
                subtype = self.viper.EqCmp(tuple_arg, arg_lit, position, info)
            else:
                subtype = self._issubtype(tuple_arg, arg_lit, ctx)
            implication = self.viper.Implies(i_in_bounds, subtype, position,
                                             info)
            trigger = self.viper.Trigger([tuple_arg], position, info)
            forall = self.viper.Forall([i_decl], [trigger], implication,
                                       position, info)
            return forall

        supertype_func = self.translate_type_literal(type, position, ctx,
                                                     alias=type_func)

        var_sub = self.viper.LocalVar('sub', self.type_type(),
                                      self.no_position(ctx), self.no_info(ctx))
        var_super = self.viper.LocalVar('super', self.type_type(),
                                        self.no_position(ctx),
                                        self.no_info(ctx))
        if concrete:
            result = self.viper.EqCmp(type_func, supertype_func, position, info)
        else:
            result = self._issubtype(type_func, supertype_func, ctx,
                                     position=position)
        return result

    def type_check(self, lhs: 'Expr', type: 'PythonType',
                   position: 'silver.ast.Position',
                   ctx: Context, concrete: bool = False) -> Expr:
        """
        Creates an expression checking if the type of the expression ``lhs``
        is a subtype of (or, if ``concrete`` is true, equivalent to) the given
        ``type``.
        """
        info = self.no_info(ctx)
        type_func = self.typeof(lhs, ctx)
        if isinstance(type, OptionalType):
            # Shortcut for performance reasons: Instead of using a real union type,
            # we say that lhs is None or it has the given type.
            none_part = self.viper.EqCmp(lhs, self.viper.NullLit(position, info),
                                         position, info)
            just_part = self.subtype_check(type_func, type.cls, position, ctx,
                                           concrete=concrete)
            return self.viper.Or(none_part, just_part, position, info)
        return self.subtype_check(type_func, type, position, ctx,
                                  concrete=concrete)

    def translate_type_literal(self, type: 'PythonType', position: 'Position',
                               ctx: Context, alias: Expr = None) -> Expr:
        """
        Translates the given type to a type literal. If the given type is
        a generic type with missing type argument information, the type
        arguments will be taken from the ``alias`` expression (which typically
        would be some expression known to be equivalent to the type being
        described).
        """
        info = self.no_info(ctx)
        if isinstance(type, TypeVar):
            if type.target_type:
                # Type parameter belongs to generic class.
                key = (type.target_type.name, type.name)
            else:
                # Type parameter belongs to generic method.
                key = (type.name,)
            return ctx.bound_type_vars[key]
        if type is None:
            type = ctx.module.global_module.classes['NoneType']
        args = []
        if isinstance(type, GenericType):
            for arg in type.type_args:
                args.append(self.translate_type_literal(arg, position, ctx))
            if not isinstance(type, UnionType):
                type = type.cls
        elif isinstance(type, PythonClass) and type.type_vars:
            assert alias
            for index, arg in enumerate(type.type_vars):
                args.append(self.get_type_arg(alias, type, index, ctx))
        if type.name == TUPLE_TYPE:
            if args:
                seq_arg = self.viper.ExplicitSeq(args, position, info)
            else:
                seq_arg = self.viper.EmptySeq(self.type_type(), position, info)
            args = [seq_arg]
        if type.name == 'Union':
            type_func_name = 'union_type_' + str(len(args))
        else:
            type_func_name = type.sil_name
        type_func = self.viper.DomainFuncApp(type_func_name, args,
                                             self.type_type(), position,
                                             info, self.type_domain)
        return type_func

    def get_type_arg(self, type_expr: Expr, target_type: PythonType,
                     index: int, ctx: Context) -> Expr:
        """
        Returns an expression denoting the type argument of the given type
        expression ``type_expr`` relative to the given ``target_type`` with
        the index ``index``.
        """
        arg_func_name = target_type.sil_name + '_arg'
        index_lit = self.viper.IntLit(index, self.no_position(ctx),
                                      self.no_info(ctx))
        args = [type_expr, index_lit]
        result = self.viper.DomainFuncApp(arg_func_name, args, self.type_type(),
                                          self.no_position(ctx),
                                          self.no_info(ctx),
                                          self.type_domain)
        return result

    def get_ref_type_arg(self, target_expr: Expr, target_type: PythonType,
                         index: int, ctx: Context):
        typeof = self.typeof(target_expr, ctx)
        return self.get_type_arg(typeof, target_type, index, ctx)

    def type_comp(self, lhs: Expr, rhs: Expr, ctx: Context, eq=False) -> Expr:
        """
        Creates an expression of the from 'typeof(lhs) == typeof(rhs)'.
        """
        type_func_lhs = self.typeof(lhs, ctx)
        type_func_rhs = self.typeof(rhs, ctx)
        if eq:
            return self.viper.EqCmp(type_func_lhs, type_func_rhs,
                                    self.no_position(ctx), self.no_info(ctx))
        else:
            return self.viper.NeCmp(type_func_lhs, type_func_rhs,
                                    self.no_position(ctx), self.no_info(ctx))



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""``MustInvoke`` obligation implementation."""


import ast

from typing import Any, Dict, List, Optional

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonMethod,
)
from nagini_translation.translators.obligation.inexhale import (
    InexhaleObligationInstanceMixin,
    ObligationInhaleExhale,
)
from nagini_translation.translators.obligation.types.base import (
    ObligationInstance,
    Obligation,
)


_OBLIGATION_NAME = 'MustInvoke'
_BOUNDED_PREDICATE_NAME = _OBLIGATION_NAME + 'Bounded'
_UNBOUNDED_PREDICATE_NAME = _OBLIGATION_NAME + 'Unbounded'
_CREDIT_PREDICATE_NAME = _OBLIGATION_NAME + 'Credit'


class MustInvokeObligationInstance(
        InexhaleObligationInstanceMixin, ObligationInstance):
    """Class representing instance of ``MustInvoke`` obligation."""

    def __init__(
            self, obligation: 'MustInvokeObligation', node: ast.expr,
            measure: Optional[ast.expr], target: ast.expr) -> None:
        super().__init__(obligation, node)
        self._measure = measure
        self._target = target

    def _get_inexhale(self, ctx: Context) -> ObligationInhaleExhale:
        return ObligationInhaleExhale(
            sil.PredicateAccess(_BOUNDED_PREDICATE_NAME, self.get_target()),
            sil.PredicateAccess(_UNBOUNDED_PREDICATE_NAME, self.get_target()),
            sil.PredicateAccess(_CREDIT_PREDICATE_NAME, self.get_target()),
            credit_only=obligation_config.disable_must_invoke)

    def is_fresh(self) -> bool:
        return self._measure is None

    def get_measure(self) -> sil.IntExpression:
        assert not self.is_fresh()
        return sil.PythonIntExpression(self._measure)

    def get_target(self) -> sil.RefExpression:
        return sil.PythonRefExpression(self._target)


class MustInvokeObligation(Obligation):
    """Class representing ``MustInvoke`` obligation."""

    def __init__(self) -> None:
        if obligation_config.disable_must_invoke:
            super().__init__([_CREDIT_PREDICATE_NAME], [])
        else:
            super().__init__([
                _BOUNDED_PREDICATE_NAME,
                _UNBOUNDED_PREDICATE_NAME,
                _CREDIT_PREDICATE_NAME], [])

    def identifier(self) -> str:
        return _OBLIGATION_NAME

    def check_node(
            self, node: ast.Call,
            obligation_info: 'PythonMethodObligationInfo',
            method: PythonMethod) -> Optional[MustInvokeObligationInstance]:
        if (isinstance(node.func, ast.Name) and
                node.func.id == 'token'):
            target = node.args[0]
            measure = node.args[1] if len(node.args) > 1 else None
            instance = MustInvokeObligationInstance(
                self, node, measure, target)
            return instance
        else:
            return None

    def generate_axiomatized_preconditions(
            self, obligation_info: 'PythonMethodObligationInfo',
            interface_dict: Dict[str, Any]) -> List[sil.BoolExpression]:
        return []

    def create_leak_check(self, var_name: str) -> List[sil.BoolExpression]:
        return [
            self._create_predicate_for_perm(
                _BOUNDED_PREDICATE_NAME, var_name),
            self._create_predicate_for_perm(
                _UNBOUNDED_PREDICATE_NAME, var_name),
        ]

    def create_ctoken_use(self, node: ast.Call) -> sil.Acc:
        """Create a ``ctoken`` use in contract."""
        target = sil.PythonRefExpression(node.args[0])
        return sil.Acc(sil.PredicateAccess(_CREDIT_PREDICATE_NAME, target))



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.lib.constants import (
    END_LABEL,
    ERROR_NAME,
    FILE_VAR,
    GLOBAL_VAR_FIELD,
    MAIN_METHOD_NAME,
    MODULE_VARS,
    NAME_VAR,
    OBJECT_TYPE,
    PRIMITIVES,
    STRING_TYPE,
)
from nagini_translation.lib.program_nodes import (
    GenericType,
    MethodType,
    PythonExceptionHandler,
    PythonField,
    PythonMethod,
    PythonModule,
    PythonTryBlock,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    DomainFuncApp,
    Expr,
    Position,
    Stmt,
    VarDecl,
)
from nagini_translation.lib.util import (
    flatten,
    get_body_indices,
    get_surrounding_try_blocks,
    InvalidProgramException
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator
from typing import List, Tuple


class MethodTranslator(CommonTranslator):

    def get_parameter_typeof(self, param: PythonVar,
                             ctx: Context) -> 'silver.ast.DomainFuncApp':
        """
        Creates an expression checking if the given parameter has its type,
        to be assumed in preconditions and/or postconditions. If possible,
        the expression is wrapped in an InhaleExhaleExpression s.t. it is
        just assumed, not checked. Generally this seems to be possible with
        types, but not with type arg numbers, because the latter encodes length
        for tuples.
        """
        no_pos = self.no_position(ctx)
        result = self.var_type_check(param.sil_name, param.type, no_pos, ctx)
        return result

    def _translate_pres(self, method: PythonMethod,
                        ctx: Context) -> List[Expr]:
        """
        Translates the preconditions specified for 'method'.
        """
        pres = []
        for pre, aliases in method.precondition:
            with ctx.additional_aliases(aliases):
                stmt, expr = self.translate_expr(pre, ctx, self.viper.Bool, True)
            if stmt:
                raise InvalidProgramException(pre, 'purity.violated')
            pres.append(expr)

        if method.cls and method.method_type is MethodType.normal:
            error_string = '"call receiver is not None"'
            pos = self.to_position(method.node, ctx, error_string)
            not_null = self.viper.NeCmp(next(iter(method.args.values())).ref(),
                                        self.viper.NullLit(
                                            self.no_position(ctx),
                                            self.no_info(ctx)),
                                        pos,
                                        self.no_info(ctx))
            pres = [not_null] + pres

        return pres

    def _translate_posts(self, method: PythonMethod,
                         err_var: 'viper.ast.LocalVar',
                         ctx: Context) -> List[Expr]:
        """
        Translates the postconditions specified for 'method'.
        """
        ctx.obligation_context.is_translating_posts = True
        posts = []
        no_error = self.viper.EqCmp(err_var,
                                    self.viper.NullLit(self.no_position(ctx),
                                                       self.no_info(ctx)),
                                    self.no_position(ctx), self.no_info(ctx))
        for post, aliases in method.postcondition:
            with ctx.additional_aliases(aliases):
                stmt, expr = self.translate_expr(post, ctx, self.viper.Bool, True)
            if stmt:
                raise InvalidProgramException(post, 'purity.violated')
            if method.declared_exceptions:
                expr = self.viper.Implies(no_error, expr,
                                          self.to_position(post, ctx),
                                          self.no_info(ctx))
            posts.append(expr)

        ctx.obligation_context.is_translating_posts = False
        return posts

    def _translate_exceptional_posts(self, method: PythonMethod,
                                     err_var: 'viper.ast.LocalVar',
                                     ctx: Context) -> List[Expr]:
        """
        Translates the exceptional postconditions specified for 'method'.
        """
        ctx.obligation_context.is_translating_posts = True
        posts = []
        error = self.viper.NeCmp(err_var,
                                 self.viper.NullLit(self.no_position(ctx),
                                                    self.no_info(ctx)),
                                 self.no_position(ctx), self.no_info(ctx))
        error_type_conds = []
        error_string = '"method only raises exceptions of type{0} {1}"'.format(
            's' if len(method.declared_exceptions) > 1 else '',
            ', '.join([e.name for e in method.declared_exceptions]))
        error_type_pos = self.to_position(method.node, ctx, error_string)
        for exception in method.declared_exceptions:
            has_type = self.var_type_check(ERROR_NAME,
                                           exception,
                                           error_type_pos,
                                           ctx, inhale_exhale=False)
            error_type_conds.append(has_type)
            condition = self.viper.And(error, has_type, self.no_position(ctx),
                                       self.no_info(ctx))
            assert ctx.current_contract_exception is None
            ctx.current_contract_exception = exception
            for post, aliases in method.declared_exceptions[exception]:
                with ctx.additional_aliases(aliases):
                    stmt, expr = self.translate_expr(post, ctx, self.viper.Bool, True)
                if stmt:
                    raise InvalidProgramException(post, 'purity.violated')
                expr = self.viper.Implies(condition, expr,
                                          self.to_position(post, ctx),
                                          self.no_info(ctx))
                posts.append(expr)
            ctx.current_contract_exception = None

        error_type_cond = None
        for type in error_type_conds:
            if error_type_cond is None:
                error_type_cond = type
            else:
                error_type_cond = self.viper.Or(error_type_cond, type,
                                                error_type_pos,
                                                self.no_info(ctx))
        if error_type_cond is not None:
            posts.append(self.viper.Implies(error, error_type_cond,
                                            error_type_pos,
                                            self.no_info(ctx)))

        ctx.obligation_context.is_translating_posts = False
        return posts

    def _create_typeof_pres(self, func: PythonMethod, is_constructor: bool,
                            ctx: Context) -> List[DomainFuncApp]:
        """
        Creates 'typeof' preconditions for function arguments.
        """
        args = func.get_args()
        pres = []
        for i, arg in enumerate(args):
            if not (arg.type.name in PRIMITIVES):
                if i == 0:
                    if is_constructor:
                        continue
                    if func.method_type == MethodType.class_method:
                        cls_arg = arg.ref()
                        type_check = self.type_factory.subtype_check(
                            cls_arg, func.cls, self.no_position(ctx), ctx)
                        pres.append(type_check)
                        continue
                type_check = self.get_parameter_typeof(arg, ctx)
                pres.append(type_check)
        if func.var_arg:
            type_check = self.get_parameter_typeof(func.var_arg, ctx)
            pres.append(type_check)
        pos = self.no_position(ctx)
        info = self.no_info(ctx)
        if func.kw_arg:
            type_check = self.get_parameter_typeof(func.kw_arg, ctx)
            set_ref = self.viper.SetType(self.viper.Ref)
            dict_acc_field = self.viper.Field('dict_acc', set_ref,
                                              pos, info)
            field_acc = self.viper.FieldAccess(func.kw_arg.ref(),
                                               dict_acc_field,
                                               pos, info)
            full_perm = self.viper.FullPerm(self.no_position(ctx),
                                            self.no_info(ctx))
            acc_pred = self.viper.FieldAccessPredicate(field_acc, full_perm,
                                                       pos, info)
            pres.append(type_check)
            pres.append(acc_pred)
        if func.cls:
            # Add upper bound information for type variables.
            for name, var in func.cls.type_vars.items():
                var_expr = self.type_factory.translate_type_literal(var, pos,
                                                                    ctx)
                check = self.type_factory.subtype_check(var_expr, var.bound,
                                                        pos, ctx)
                pres.append(check)
        if func.name != 'Eval':
            # Make an exception for Eval because it uses type variables in ways we don't
            # generally support.
            for name, var in func.type_vars.items():
                check = self.type_factory.subtype_check(var.type_expr, var.bound,
                                                        pos, ctx)
                pres.append(check)
        return pres

    def _translate_params(self, func: PythonMethod,
                          ctx: Context) -> List[VarDecl]:
        args = []
        for arg in func.get_args():
            args.append(arg.decl)
        if func.var_arg:
            args.append(func.var_arg.decl)
        if func.kw_arg:
            args.append(func.kw_arg.decl)
        return args

    def translate_function(self, func: PythonMethod,
                           ctx: Context) -> 'silver.ast.Function':
        """
        Translates a pure Python function (may or not belong to a class) to a
        Viper function
        """
        old_function = ctx.current_function
        ctx.current_function = func
        self.bind_type_vars(func, ctx)
        pos = self.to_position(func.node, ctx)
        if not func.type:
            raise InvalidProgramException(func.node, 'function.type.none')
        type = self.translate_type(func.type, ctx)
        args = self._translate_params(func, ctx)
        if func.declared_exceptions:
            raise InvalidProgramException(func.node,
                                          'function.throws.exception')
        # Create preconditions
        pres = self._translate_pres(func, ctx)
        # Create postconditions
        posts = []
        for post, aliases in func.postcondition:
            with ctx.additional_aliases(aliases):
                stmt, expr = self.translate_expr(post, ctx, self.viper.Bool)
            if stmt:
                raise InvalidProgramException(post, 'purity.violated')
            posts.append(expr)
        # Create typeof preconditions
        pres = self._create_typeof_pres(func, False, ctx) + pres
        if func.type.name not in PRIMITIVES:
            res_type_pos = self.to_position(func.node, ctx,
                                            '"return type is correct"')
            res_type = self.translate_type(func.type, ctx)
            result = self.viper.Result(res_type, res_type_pos, self.no_info(ctx))
            check = self.type_check(result, func.type, res_type_pos, ctx)
            posts = [check] + posts

        statements = func.node.body
        start, end = get_body_indices(statements)
        # Translate body
        actual_body = statements[start:end]
        if (func.contract_only or
                (len(actual_body) == 1 and isinstance(actual_body[0], ast.Expr) and
                 isinstance(actual_body[0].value, ast.Ellipsis))):
            body = None
        else:
            body = self.translate_exprs(actual_body, func, ctx)
        ctx.current_function = old_function
        name = func.sil_name
        return self.viper.Function(name, args, type, pres, posts, body,
                                   pos, self.no_info(ctx))

    def extract_contract(self, method: PythonMethod, errorvarname: str,
                         is_constructor: bool,
                         ctx: Context) -> Tuple[List[Expr], List[Expr]]:
        """
        Extracts the pre and postcondition from a given method
        """
        error_var_ref = self.viper.LocalVar(errorvarname, self.viper.Ref,
                                            self.no_position(ctx),
                                            self.no_info(ctx))
        # Create preconditions
        pres = self._translate_pres(method, ctx)
        # Create postconditions
        posts = self._translate_posts(method, error_var_ref, ctx)
        # Create exceptional postconditions
        posts += self._translate_exceptional_posts(method, error_var_ref, ctx)
        # Create typeof preconditions
        type_pres = self._create_typeof_pres(method, is_constructor, ctx)
        pres = type_pres + pres

        result_post = self._create_result_type_post(method, error_var_ref, ctx)
        posts = result_post + posts
        return pres, posts

    def _create_result_type_post(self, method: PythonMethod, error_var_ref,
                                 ctx: Context) -> List[Expr]:
        if method.type and method.type.name not in PRIMITIVES:
            result = self._create_single_result_post(method, error_var_ref,
                ctx.result_var.ref(method.node, ctx), ctx)
            return result
        else:
            return []

    def _create_single_result_post(self, method: PythonMethod, error_var_ref,
                                   result_var, ctx: Context) -> List[Expr]:
        no_pos = self.no_position(ctx)
        method_pos = self.to_position(method.node, ctx,
                                      '"return type is correct"')
        no_info = self.no_info(ctx)

        check = self.type_check(result_var,
                                method.type, method_pos, ctx)
        if method.declared_exceptions:
            no_error = self.viper.EqCmp(error_var_ref,
                                        self.viper.NullLit(no_pos, no_info),
                                        method_pos, no_info)
            check = self.viper.Implies(no_error, check, method_pos, no_info)
        result = [check]
        return result

    def get_all_field_accs(self, fields: List['silver.ast.Field'],
            self_var: 'silver.ast.LocalVar', position: 'silver.ast.Position',
            ctx: Context) -> List['silver.ast.FieldAccessPredicate']:
        """
        Creates access predicates for all fields in fields.
        """
        accs = []
        for field in fields:
            acc = self.viper.FieldAccess(self_var, field,
                                         position, self.no_info(ctx))
            perm = self.viper.FullPerm(position, self.no_info(ctx))
            pred = self.viper.FieldAccessPredicate(acc, perm, position,
                                                   self.no_info(ctx))
            accs.append(pred)
        return accs

    def _create_method_epilog(self, method: PythonMethod,
                              ctx: Context) -> List[Stmt]:
        """
        Hook to generate the method epilog.
        """
        end_name = ctx.get_label_name(END_LABEL)
        return [self.viper.Label(end_name, self.no_position(ctx),
                                 self.no_info(ctx))]

    def _create_init_pres(self, method: PythonMethod,
                          ctx: Context) -> List[Expr]:
        """
        Generates preconditions specific to the '__init__' method.
        """
        self_var = method.args[next(iter(method.args))].ref()
        fields = method.cls.all_fields
        accs = [self.get_may_set_predicate(self_var, f, ctx) for f in fields]
        null = self.viper.NullLit(self.no_position(ctx), self.no_info(ctx))
        not_null = self.viper.NeCmp(self_var, null, self.no_position(ctx),
                                    self.no_info(ctx))

        return [not_null] + accs

    def _create_local_vars_for_params(self, method: PythonMethod,
                                      ctx: Context) -> List[Stmt]:
        """Creates LocalVarAssigns for each parameter."""
        assign_stmts = []
        for name, arg in method.args.items():
            arg_var = ctx.current_function.create_variable(name, arg.type,
                                                           self.translator)
            arg_assign = self.viper.LocalVarAssign(arg_var.ref(), arg.ref(),
                                                   self.no_position(ctx),
                                                   self.no_info(ctx))
            assign_stmts.append(arg_assign)
            ctx.set_alias(name, arg_var, arg)

        return assign_stmts

    def bind_type_vars(self, method: PythonMethod, ctx: Context) -> None:
        """
        Binds the names of type variables of the given method and its class
        and superclasses to expressions denoting the values of said types.
        """
        ctx.bound_type_vars = {}
        # Class type variables first.
        if method.cls and method.method_type is MethodType.normal:
            cls = method.cls
            while cls:
                for name, var in cls.type_vars.items():
                    self_arg = next(iter(method.args.values())).ref()
                    literal = self.type_factory.get_ref_type_arg(self_arg,
                                                                 var.target_type,
                                                                 var.index, ctx)
                    ctx.bound_type_vars[(var.target_type.name, name)] = literal
                cls = cls.superclass
                if isinstance(cls, GenericType):
                    cls = cls.cls
        # Now method type variables.
        for name, var in method.type_vars.items():
            if callable(var.type_expr):
                # var.type_expr is currently the function that will create the
                # type expression when given the parameter that defines the
                # type variable (identified by var.target_node).
                for i, python_var in enumerate(method.args.values()):
                    if python_var.node is var.target_node:
                        ref = python_var.ref()
                typeof = self.type_factory.typeof(ref, ctx)
                literal = var.type_expr(self.type_factory, typeof, ctx)
                var.type_expr = literal
            else:
                literal = var.type_expr
            ctx.bound_type_vars[(var.name,)] = literal
        if method.name == 'Eval':
            # Eval uses type parameters in ways we don't usually support;
            # while translating it, we're treating the type parameter V as type
            # object.
            object_type = ctx.module.global_module.classes[OBJECT_TYPE]
            literal = self.type_factory.translate_type_literal(object_type,
                                                               self.no_position(ctx), ctx)
            ctx.bound_type_vars[('V',)] = literal

    def _check_self_type(self, method: PythonMethod, ctx: Context) -> Stmt:
        """
        Return an statement checking the type of the 'self' variable (only for methods in classes).
        """
        type_check = self.type_factory.type_check(
                next(iter(method.args.values())).ref(), method.cls, self.no_position(ctx), ctx,
                concrete=True)
        return self.viper.Inhale(type_check, self.no_position(ctx),
                                            self.no_info(ctx))

    def _translate_method_body(self, method: PythonMethod, ctx: Context) -> List[Stmt]:
        body = []
        statements = method.node.body
        body_start, body_end = get_body_indices(statements)
        # Create local variables for parameters
        body.extend(self._create_local_vars_for_params(method, ctx))
        body += flatten(
            [self.translate_stmt(stmt, ctx) for stmt in
                method.node.body[body_start:body_end]])
        return body

    def _translate_try_handlers(self, method: PythonMethod, ctx: Context) -> List[Stmt]:
        """
        Translates the handlers of all try blocks, as well as finally blocks.
        """
        stmts = []
        for block in method.try_blocks:
            for handler in block.handlers:
                stmts += self.translate_handler(handler, ctx)
            if block.else_block:
                stmts += self.translate_handler(block.else_block, ctx)
            if block.finally_block or block.with_item:
                stmts += self.translate_finally(block, ctx)
        return stmts

    def _method_body_postamble(self, method: PythonMethod, ctx: Context) -> List[Stmt]:
        postamble = []
        end_label = ctx.get_label_name(END_LABEL)
        postamble.append(self.viper.Goto(end_label, self.no_position(ctx), self.no_info(ctx)))
        assert not ctx.var_aliases
        postamble += self._translate_try_handlers(method, ctx)
        postamble += self.add_handlers_for_inlines(ctx)
        return postamble

    def create_method_error_var(self, ctx: Context) -> PythonVar:
        error_var = PythonVar(ERROR_NAME, None,
                              ctx.module.global_module.classes['Exception'])
        error_var.process(ERROR_NAME, self.translator)
        return error_var

    def translate_method(self, method: PythonMethod,
                         ctx: Context) -> 'silver.ast.Method':
        """
        Translates an impure Python function (may or not belong to a class) to
        a Viper method
        """
        old_function = ctx.current_function
        ctx.current_function = method
        args = self._translate_params(method, ctx)
        self.bind_type_vars(method, ctx)

        results = [res.decl for res in method.get_results()]
        error_var = self.create_method_error_var(ctx)
        error_var_decl = error_var.decl
        error_var_ref = error_var.ref()
        method.error_var = error_var
        pres, posts = self.extract_contract(method, ERROR_NAME, False, ctx)
        if method.cls and method.name == '__init__':
            init_pres = self._create_init_pres(method, ctx)
            pres = init_pres + pres
        if method.declared_exceptions:
            results.append(error_var_decl)

        # Translate body
        body = []
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        if method.cls and method.method_type == MethodType.normal:
            body.append(self._check_self_type(method, ctx))
        if method.type:
            # Assign null as the default return value to the return variable.
            assign_none = self.viper.LocalVarAssign(method.result.ref(),
                                                    self.viper.NullLit(no_pos,
                                                                       no_info),
                                                    no_pos, no_info)
            body.append(assign_none)
        if method.contract_only:
            false = self.viper.FalseLit(self.no_position(ctx),
                                        self.no_info(ctx))
            assume_false = self.viper.Inhale(false, self.no_position(ctx),
                                             self.no_info(ctx))
            body.append(assume_false)
            locals = []
        else:
            body.append(self.viper.LocalVarAssign(error_var_ref,
                self.viper.NullLit(self.no_position(ctx),
                                    self.no_info(ctx)),
                self.no_position(ctx), self.no_info(ctx)))
            if method.declared_exceptions:
                locals = []
            else:
                locals = [error_var_decl]
            body += self._translate_method_body(method, ctx)
            for arg in method.get_args():
                ctx.remove_alias(arg.name)
            body += self._method_body_postamble(method, ctx)
            locals += [local.decl for local in method.get_locals()
                       if not local.name.startswith('lambda')]
            body += self._create_method_epilog(method, ctx)
        name = method.sil_name
        nodes = self.create_method_node(
            ctx, name, args, results, pres, posts, locals, body,
            self.to_position(method.node, ctx), self.no_info(ctx),
            method=method)
        ctx.current_function = old_function
        return nodes

    def _assign_exit_vars(self, block: PythonTryBlock, type_var: PythonVar,
                          value_var: PythonVar, traceback_var: PythonVar,
                          pos: Position, ctx: Context) -> Stmt:
        """
        Assigns the exception and its type to the given variables if there is
        an uncaught exception, otherwise assigns None to everything.
        """
        info = self.no_info(ctx)
        null = self.viper.NullLit(pos, info)
        one = self.viper.IntLit(1, pos, info)
        code_var = block.get_finally_var(self.translator)
        error_cond = self.viper.GtCmp(code_var.ref(), one, pos, info)
        error_case = []
        no_error_case = []
        # FIXME: Cannot currently assign None to type variable, because types
        # aren't objects.
        for var in [value_var, traceback_var]:
            assign = self.viper.LocalVarAssign(var.ref(), null, pos, info)
            no_error_case.append(assign)

        value_assign = self.viper.LocalVarAssign(value_var.ref(),
                                                 block.error_var.ref(), pos,
                                                 info)
        error_case.append(value_assign)
        error_type = self.type_factory.typeof(block.error_var.ref(), ctx)
        type_assign = self.viper.LocalVarAssign(type_var.ref(), error_type,
                                                pos, info)
        error_case.append(type_assign)

        tb_class = ctx.module.global_module.classes['traceback']
        tb_type = self.type_check(traceback_var.ref(), tb_class, pos, ctx)
        inhale_types = self.viper.Inhale(tb_type, pos, info)
        error_case.append(inhale_types)
        then_block = self.translate_block(error_case, pos, info)
        else_block = self.translate_block(no_error_case, pos, info)
        return self.viper.If(error_cond, then_block, else_block, pos, info)

    def translate_finally(self, block: PythonTryBlock,
                          ctx: Context) -> List[Stmt]:
        """
        Creates a code block representing the finally-block belonging to block,
        to be put at the end of a Viper method.
        """
        pos = self.to_position(block.node, ctx)
        info = self.no_info(ctx)
        label_name = ctx.get_label_name(block.finally_name)
        label = self.viper.Label(label_name,
                                 self.to_position(block.node, ctx),
                                 self.no_info(ctx))
        body = [label]
        if block.finally_block:
            for stmt in block.finally_block:
                body += self.translate_stmt(stmt, ctx)
        else:
            # With-block
            ctx_type = self.get_type(block.with_item.context_expr, ctx)
            ctx_var = block.with_var
            exit_type = ctx_type.get_method('__exit__').type
            exit_res = ctx.current_function.create_variable('exit_res',
                                                            exit_type,
                                                            self.translator)

            type_class = ctx.module.global_module.classes['type']
            exception_class = ctx.module.global_module.classes['Exception']
            tb_class = ctx.module.global_module.classes['traceback']
            # The __exit__ method takes three arguments: type, value and
            # traceback.
            type_var = ctx.actual_function.create_variable('t', type_class,
                                                           self.translator)
            value_var = ctx.actual_function.create_variable('e',
                                                            exception_class,
                                                            self.translator)
            traceback_var = ctx.actual_function.create_variable('tb',
                                                                tb_class,
                                                                self.translator)
            body.append(self._assign_exit_vars(block, type_var, value_var,
                                               traceback_var, pos, ctx))
            exit_call = self.get_method_call(ctx_type, '__exit__',
                                             [ctx_var.ref(), type_var.ref(),
                                              value_var.ref(),
                                              traceback_var.ref()],
                                             [ctx_type, None, None, None],
                                             [exit_res.ref()],
                                             block.with_item.context_expr, ctx)
            body.extend(exit_call)
        finally_var = block.get_finally_var(self.translator)
        if finally_var.sil_name in ctx.var_aliases:
            finally_var = ctx.var_aliases[finally_var.sil_name]
        tries = get_surrounding_try_blocks(ctx.actual_function.try_blocks,
                                           block.node)
        post_label = ctx.get_label_name(block.post_name)
        goto_post = self.viper.Goto(post_label, pos, info)
        end_label = ctx.get_label_name(END_LABEL)
        goto_end = self.viper.Goto(end_label, pos, info)
        empty_stmt = self.translate_block([], pos, info)

        except_block = []
        return_block = []
        for current in tries:
            if not return_block and current.finally_block:
                # Propagate return value
                var_next = current.get_finally_var(self.translator)
                if var_next.sil_name in ctx.var_aliases:
                    var_next = ctx.var_aliases[var_next.sil_name]
                next_assign = self.viper.LocalVarAssign(var_next.ref(),
                                                        finally_var.ref(),
                                                        pos, info)
                # Goto finally block
                next_label = ctx.get_label_name(current.finally_name)
                goto_next = self.viper.Goto(next_label, pos, info)
                return_block = [next_assign, goto_next]
            for handler in current.handlers:
                # If handler applies,
                # goto handler
                err_var = block.get_error_var(self.translator)
                if err_var.sil_name in ctx.var_aliases:
                    err_var = ctx.var_aliases[err_var.sil_name]
                condition = self.var_type_check(err_var.sil_name,
                                                handler.exception,
                                                self.to_position(handler.node,
                                                                 ctx),
                                                ctx, inhale_exhale=False)
                label_name = ctx.get_label_name(handler.name)
                goto = self.viper.Goto(label_name, pos, info)
                if_handler = self.viper.If(condition, goto, empty_stmt, pos,
                                           info)
                except_block.append(if_handler)
            if current.finally_block:
                # Propagate return value
                # Goto finally block
                except_block += return_block
                break
        if not return_block:
            return_block = [goto_end]
        if ctx.actual_function.declared_exceptions:
            # Assign error to error output var
            error_var = ctx.error_var.ref()
            block_error_var = block.get_error_var(self.translator)
            if block_error_var.sil_name in ctx.var_aliases:
                block_error_var = ctx.var_aliases[block_error_var.sil_name]
            assign = self.viper.LocalVarAssign(
                error_var, block_error_var.ref(), pos, info)
            except_block.append(assign)
            except_block.append(goto_end)
        else:
            error_string = '"method raises no exceptions"'
            error_pos = self.to_position(block.node, ctx, error_string)
            false = self.viper.FalseLit(error_pos, info)
            assert_false = self.viper.Exhale(false, error_pos, info)
            except_block.append(assert_false)
            except_block.append(goto_end)
        except_block = self.translate_block(except_block, pos, info)
        return_block = self.translate_block(return_block, pos, info)

        number_zero = self.viper.IntLit(0, pos, info)
        greater_zero = self.viper.GtCmp(finally_var.ref(), number_zero, pos,
                                        info)
        number_one = self.viper.IntLit(1, pos, info)
        greater_one = self.viper.GtCmp(finally_var.ref(), number_one, pos, info)
        if_return = self.viper.If(greater_zero, return_block, goto_post, pos,
                                  info)
        if_except = self.viper.If(greater_one, except_block, if_return, pos,
                                  info)
        body += [if_except]
        return body

    def translate_handler(self, handler: PythonExceptionHandler,
                          ctx: Context) -> List[Stmt]:
        """
        Creates a code block representing an exception handler, to be put at
        the end of a Viper method
        """
        label_name = ctx.get_label_name(handler.name)
        label = self.viper.Label(label_name,
                                 self.to_position(handler.node, ctx),
                                 self.no_info(ctx))
        old_var_aliases = ctx.var_aliases
        ctx.var_aliases = handler.try_block.handler_aliases
        no_position = self.no_position(ctx)
        no_info = self.no_info(ctx)
        body = []
        if handler.exception_name:
            err_var = handler.try_block.get_error_var(self.translator)
            if err_var.sil_name in ctx.var_aliases:
                err_var = ctx.var_aliases[err_var.sil_name]
            ctx.var_aliases[handler.exception_name] = err_var
            err_var.type = handler.exception

            body.append(self.set_var_defined(err_var, no_position, no_info))
        for stmt in handler.body:
            body += self.translate_stmt(stmt, ctx)
        body_block = self.translate_block(body,
                                          self.to_position(handler.node, ctx),
                                          no_info)
        if handler.try_block.finally_block:
            next = handler.try_block.finally_name
            finally_var = handler.try_block.get_finally_var(self.translator)
            if finally_var.sil_name in ctx.var_aliases:
                finally_var = ctx.var_aliases[finally_var.sil_name]
            lhs = finally_var.ref()
            rhs = self.viper.IntLit(0, no_position, no_info)
            var_set = self.viper.LocalVarAssign(lhs, rhs, no_position, no_info)
            next_var_set = [var_set]
        else:
            next = 'post_' + handler.try_block.name
            next_var_set = []
        label_name = ctx.get_label_name(next)
        goto_end = self.viper.Goto(label_name,
                                   self.to_position(handler.node, ctx),
                                   no_info)
        ctx.var_aliases = old_var_aliases
        return [label, body_block] + next_var_set + [goto_end]

    def _initialize_module(self, module: PythonModule, ctx: Context) -> None:
        """
        Creates the variables representing the built-in global variables that exist
        in each module.
        """
        if isinstance(module.names_var, str):
            names_decl = self.viper.LocalVarDecl(module.names_var,
                                                 self.viper.SetType(self.name_type()),
                                                 self.no_position(ctx), self.no_info(ctx))
            names_ref = self.viper.LocalVar(module.names_var,
                                            self.viper.SetType(self.name_type()),
                                            self.no_position(ctx), self.no_info(ctx))
            def_decl = self.viper.LocalVarDecl(module.defined_var, self.viper.Bool,
                                               self.no_position(ctx), self.no_info(ctx))
            def_ref = self.viper.LocalVar(module.defined_var, self.viper.Bool,
                                          self.no_position(ctx), self.no_info(ctx))
            module.names_var = (names_decl, names_ref)
            module.defined_var = (def_decl, def_ref)

    def _initialize_main_state(self, modules: List[PythonModule], main: PythonModule,
                               ctx: Context) -> Tuple[List['silver.ast.LocalVarDecl'],
                                                      List[Stmt]]:
        """
        Generates a list of statements that initialize the local state used in the main
        method. This includes inhaling permissions to mutable global variables and setting
        up the set of known names for each module.
        """
        stmts = []
        locals = []
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        false_lit = self.viper.FalseLit(no_pos, no_info)
        true_lit = self.viper.TrueLit(no_pos, no_info)
        empty_set = self.viper.EmptySet(self.name_type(), no_pos, no_info)
        global_field = self.viper.Field(GLOBAL_VAR_FIELD, self.viper.Ref, no_pos, no_info)
        full_perm = self.viper.FullPerm(no_pos, no_info)
        ninety_nine = self.viper.IntLit(99, no_pos, no_info)
        hundred = self.viper.IntLit(100, no_pos, no_info)
        # For module variables like __file__, a (rather arbitrary) permission amount of
        # 99% is available. The intention is that they should not be changed, and the
        # natural way to enforce this is to give users less than a full permission.
        part_perm = self.viper.FractionalPerm(ninety_nine, hundred, no_pos, no_info)
        for module in modules:
            if module.global_module is module:
                continue
            self._initialize_module(module, ctx)
            if module is main:
                stmts.append(self.viper.LocalVarAssign(module.defined_var[1],
                                                       true_lit, no_pos, no_info))
            else:
                stmts.append(self.viper.LocalVarAssign(module.defined_var[1],
                                                       false_lit, no_pos, no_info))
            stmts.append(self.viper.LocalVarAssign(module.names_var[1],
                                                   empty_set, no_pos, no_info))
            locals.append(module.defined_var[0])
            locals.append(module.names_var[0])
            for var in module.global_vars.values():
                if var.module is not module:
                    continue
                if var.is_final:
                    continue
                perm = full_perm
                if var.name in MODULE_VARS:
                    stmts.append(self.set_global_defined(var, module, None, ctx))
                    perm = part_perm
                var_type = self.translate_type(var.type, ctx)
                var_func = self.viper.FuncApp(var.sil_name, [], no_pos,
                                              no_info, var_type, [])
                field_access = self.viper.FieldAccess(var_func, global_field, no_pos,
                                                      no_info)
                field_pred = self.viper.FieldAccessPredicate(field_access, perm,
                                                             no_pos, no_info)
                field_type_check = self.type_check(field_access, var.type, no_pos, ctx)
                field_pred = self.viper.And(field_pred, field_type_check, no_pos, no_info)
                if var.name in MODULE_VARS:
                    var_type = self.type_check(field_access, var.type, no_pos, ctx, False)
                    field_pred = self.viper.And(field_pred, var_type, no_pos, no_info)
                if var.name == NAME_VAR:
                    main_str = self.translate_string('__main__', None, ctx)
                    str_type = ctx.module.global_module.classes[STRING_TYPE]
                    func_name = '__eq__'
                    call = self.get_function_call(str_type, func_name,
                                                  [main_str, field_access],
                                                  [None, None], var.node, ctx)
                    if module is not main:
                        call = self.viper.Not(call, no_pos, no_info)
                    field_pred = self.viper.And(field_pred, call, no_pos, no_info)
                stmts.append(self.viper.Inhale(field_pred, no_pos, no_info))
        return locals, stmts

    def _get_main_module(self, modules: List[PythonModule]) -> PythonModule:
        return [m for m in modules if m.type_prefix == '__main__'][0]

    def _create_main_method_setup(self, modules: List[PythonModule],
                                  ctx: Context) -> Tuple[PythonMethod, List[VarDecl], List[Stmt]]:
        main = self._get_main_module(modules)
        locals, init_stmts = self._initialize_main_state(modules, main, ctx)

        # Create artificial main PythonMethod that contains the execution of global
        # statements.
        ctx.current_class = None
        method_name = ctx.module.get_fresh_name('main')
        main_method = PythonMethod(MAIN_METHOD_NAME, main, None, main, False, False,
                                   main.node_factory)
        main_method._module = main
        ctx.current_function = main_method
        ctx.current_function.try_blocks = main.try_blocks
        ctx.current_function.labels = main.labels
        ctx.current_function.precondition = main.precondition
        ctx.current_function.postcondition = main.postcondition
        ctx.current_function.loop_invariants = main.loop_invariants
        ctx.current_function.process(method_name, self.translator)
        ctx.module = main
        return main_method, locals, init_stmts

    def translate_main_method(self, modules: List[PythonModule],
                              ctx: Context) -> List['silver.ast.Method']:
        """
        Translates the global statements of the program to a single method.
        """
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)

        main = self._get_main_module(modules)
        main_method, locals, stmts = self._create_main_method_setup(modules, ctx)
        method_name = main_method.sil_name

        # Translate statements in main module. When an import statement is encountered,
        # the translation will include executing the statements in the imported module.
        for stmt in main.node.body:
            stmts.extend(self.translate_stmt(stmt, ctx))

        stmts += self._method_body_postamble(main_method, ctx)
        stmts += self._create_method_epilog(main_method, ctx)

        main_locals = [local.decl for local in main_method.get_locals()
                       if not local.name.startswith('lambda')]
        for tb in main_method.try_blocks:
            main_locals.append(tb.error_var.decl)
            main_locals.append(tb.finally_var.decl)
        body = stmts
        res = self.create_method_node(ctx, method_name, [], [], [], [],
                                      main_locals + locals, body, no_pos,
                                      no_info, method=ctx.current_function)
        ctx.current_function = None
        return main_method, res





"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Information about obligation use in contracts."""


import abc
import ast
from typing import List

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.constants import PRIMITIVE_BOOL_TYPE
from nagini_translation.lib.context import Context
from nagini_translation.lib.guard_collectors import (
    GuardCollectingVisitor,
)
from nagini_translation.lib.program_nodes import (
    PythonMethod,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Stmt,
    Type,
)
from nagini_translation.sif.lib.viper_ast_extended import ViperASTExtended
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.measures import (
    MeasureMap,
)
from nagini_translation.translators.obligation.types.base import (
    ObligationInstance,
)

CURRENT_THREAD_NAME = '_cthread'
RESIDUE_LEVEL_METHOD_NAME = '_residue'
RESIDUE_LEVEL_LOOP_NAME = '_residue'
CURRENT_WAIT_LEVEL_NAME = '_current_wait_level'
CURRENT_WAIT_LEVEL_TARGET_NAME = '_cwl'
MEASURES_CALLER_NAME = '_caller_measures'
MEASURES_METHOD_NAME = '_method_measures'
MEASURES_LOOP_NAME = '_loop_measures'
LOOP_CHECK_BEFORE_NAME = '_loop_check_before'
LOOP_TERMINATION_FLAG_NAME = '_loop_termination_flag'
LOOP_ORIGINAL_MUST_TERMINATE_AMOUNT_NAME = '_loop_original_must_terminate'


def _create_guard_expression(parts: List[ast.AST]) -> sil.BigAnd:
    """Transform guarding sequence into the guard expression."""
    conjunction = sil.BigAnd([
        sil.PythonBoolExpression(part)
        for part in parts
    ])
    return conjunction


@PythonVar.register         # pylint: disable=no-member
class SilverVar:
    """A silver variable that has no representation in Python.

    This class is a structural subtype of ``PythonVar`` that allows to
    manage variables non-representable with ``PythonVar`` (like the ones
    that have ``Perm`` type) in the same way as all other variables.
    """

    def __init__(self, name: str, decl: 'viper_ast.LocalVarDecl',
                 ref: 'viper_ast.LocalVarRef') -> None:
        self.name = name
        self.sil_name = name
        self.decl = decl
        """A variable declaration."""

        self._ref = ref

    def ref(self, node: ast.AST = None,                         # pylint: disable=unused-argument
            ctx: Context = None) -> 'viper_ast.LocalVarRef':    # pylint: disable=unused-argument
        """A variable reference.

        Arguments are ignored.
        """
        return self._ref

    def process(self, sil_name: str, translator: 'Translator') -> None:
        """Just do nothing."""


class GuardedObligationInstance:
    """Obligation instance with its guard."""

    def __init__(
            self, guard: List[ast.AST],
            obligation_instance: ObligationInstance) -> None:
        self.guard = guard
        self.obligation_instance = obligation_instance

    def create_guard_expression(self) -> sil.BigAnd:
        """Create a conjunction representing a guard."""
        return _create_guard_expression(self.guard)


class BaseObligationInfo(GuardCollectingVisitor):
    """Info about the obligation use in loop/method contract."""

    def __init__(
            self, obligaton_manager: ObligationManager,
            method: PythonMethod) -> None:
        super().__init__()
        self._current_instance_map = None
        self._obligation_manager = obligaton_manager
        self._method = method
        self._all_instances = {}
        self._wait_level_guards = {}

    def visit_Call(self, node: ast.Call) -> None:
        for obligation in self._obligation_manager.obligations:
            obligation_instance = obligation.check_node(
                node, self, self._method)
            if obligation_instance:
                guard = self.current_guard[:]
                if (isinstance(node.func, ast.Name) and
                        node.func.id == "TerminatesSif"):
                    guard.append(node.args[0])
                instance = GuardedObligationInstance(
                    guard, obligation_instance)
                self._current_instance_map[obligation.identifier()].append(
                    instance)
                self._all_instances[node] = instance
                break
        else:
            if isinstance(node.func, ast.Name) and node.func.id == 'WaitLevel':
                self._wait_level_guards[node] = self.current_guard[:]
            else:
                super().visit_Call(node)

    def get_instance(self, node: ast.Call) -> GuardedObligationInstance:
        """Get ``GuardedObligationInstance`` represented by node."""
        return self._all_instances[node]

    @abc.abstractmethod
    def _get_must_terminate_instances(self) -> List[GuardedObligationInstance]:
        """Get all ``MustTerminate`` obligation instances."""

    @abc.abstractmethod
    def _check_must_terminate_measure_decrease(
            self, measure: sil.IntExpression) -> sil.BoolExpression:
        """Create a check if provided measure is smaller than current."""

    def create_termination_check(
            self, ignore_measures: bool) -> sil.BoolExpression:
        """Create a check if callee is going to terminate.

        This method is essentially a ``tcond`` macro as defined in
        ``MustTerminate`` documentation.
        """
        disjuncts = []
        for instance in self._get_must_terminate_instances():
            guard = instance.create_guard_expression()
            if ignore_measures:
                disjuncts.append(guard)
            else:
                measure_check = self._check_must_terminate_measure_decrease(
                    instance.obligation_instance.get_measure())
                disjuncts.append(sil.BigAnd([guard, measure_check]))
        return sil.BigOr(disjuncts)

    def get_wait_level_guard(self, node: ast.Call) -> sil.BigAnd:
        """Return a guard for the given WaitLevel expression."""
        return _create_guard_expression(self._wait_level_guards[node])

    def _create_var(
            self, name: str, class_name: str,
            translator: 'Translator', local: bool = False) -> PythonVar:
        module = self._method.module
        cls = module.global_module.classes[class_name]
        return self._method.create_variable(
            name, cls, translator, local=local)

    def _create_silver_var(
            self, name: str, translator: 'AbstractTranslator',
            typ: Type, local: bool = False) -> SilverVar:
        sil_name = self._method.get_fresh_name(name)
        if isinstance(translator.viper, ViperASTExtended):
            info = translator.viper.SIFInfo([], obligation_var=True)
        else:
            info = translator.viper.NoInfo
        decl = translator.viper.LocalVarDecl(
            sil_name, typ, translator.viper.NoPosition, info)
        ref = translator.viper.LocalVar(
            sil_name, typ, translator.viper.NoPosition,
            translator.viper.NoInfo)
        var = SilverVar(sil_name, decl, ref)
        if local:
            self._method.add_local(sil_name, var)
        return var

    def _create_perm_var(
            self, name: str, translator: 'AbstractTranslator',
            local: bool = False) -> PythonVar:
        return self._create_silver_var(
            name, translator, translator.viper.Perm, local=local)

    def _create_seq_var(
            self, name: str,
            translator: 'AbstractTranslator') -> PythonVar:
        return self._create_silver_var(
            name, translator, translator.viper.SeqType(translator.viper.Ref))

    def _create_measure_var(
            self, name: str, translator: 'AbstractTranslator',
            local: bool = False) -> PythonVar:
        typ = translator.viper.SeqType(
            translator.viper.DomainType('Measure$', {}, []))
        return self._create_silver_var(name, translator, typ, local=local)


class PythonMethodObligationInfo(BaseObligationInfo):
    """Info about the obligation use in a specific method."""

    def __init__(
            self, obligaton_manager: ObligationManager,
            method: PythonMethod, translator: 'AbstractTranslator') -> None:
        super().__init__(obligaton_manager, method)
        self._precondition_instances = {}
        self._postcondition_instances = {}
        for obligation in self._obligation_manager.obligations:
            obligation_id = obligation.identifier()
            self._precondition_instances[obligation_id] = []
            self._postcondition_instances[obligation_id] = []
        self._variables = {
            CURRENT_THREAD_NAME: self._create_var(
                CURRENT_THREAD_NAME, 'Thread', translator.translator),
            RESIDUE_LEVEL_METHOD_NAME: self._create_perm_var(
                RESIDUE_LEVEL_METHOD_NAME, translator),
            CURRENT_WAIT_LEVEL_NAME: self._create_perm_var(
                CURRENT_WAIT_LEVEL_NAME, translator),
            CURRENT_WAIT_LEVEL_TARGET_NAME: self._create_perm_var(
                CURRENT_WAIT_LEVEL_TARGET_NAME, translator),
        }
        caller_measure_map_var = self._create_measure_var(
            MEASURES_CALLER_NAME, translator)
        self.caller_measure_map = MeasureMap(caller_measure_map_var)
        method_measure_map_var = self._create_measure_var(
            MEASURES_METHOD_NAME, translator)
        self.method_measure_map = MeasureMap(
            method_measure_map_var)
        self._additional_preconditions = []
        self._additional_postconditions = []

    @property
    def current_thread_var(self) -> PythonVar:
        """Return variable that represents current thread argument."""
        return self._variables[CURRENT_THREAD_NAME]

    @property
    def residue_level(self) -> PythonVar:
        """Return variable that represents the residue level argument."""
        return self._variables[RESIDUE_LEVEL_METHOD_NAME]

    @property
    def current_wait_level(self) -> PythonVar:
        """Return variable that represents current wait level return value."""
        return self._variables[CURRENT_WAIT_LEVEL_NAME]

    @property
    def current_wait_level_target(self) -> PythonVar:
        """Return target variable to which the current wait level is assigned.

        .. note::
            In the wait level use in method calls encoding, we need to
            have the fresh variable that denotes the current thread's
            wait level after the precondition was exhaled, but before
            the postcondition is inhaled. We use an additional ghost
            return variable ``CURRENT_WAIT_LEVEL_NAME`` for this:

            .. code-block:: silver

                method foo(...) returns (..., _current_wait_level: Perm)

            However, when this method ``foo`` is called, we need a
            variable to which we can assign this return value.
            Therefore, we introduce a local variable
            ``CURRENT_WAIT_LEVEL_TARGET_NAME`` in each method for this
            purpose:

            .. code-block:: silver

                ..., _cwl := foo(...)

            Note that this variable ``CURRENT_WAIT_LEVEL_TARGET_NAME``
            is otherwise not used.
        """
        return self._variables[CURRENT_WAIT_LEVEL_TARGET_NAME]

    def traverse_contract(self) -> None:
        """Collect all needed information about obligations."""
        self._traverse_preconditions()
        self._traverse_postconditions()
        self._traverse_declared_exceptions()

    def _traverse_preconditions(self) -> None:
        """Collect all needed information about obligations."""
        assert self._current_instance_map is None
        self._current_instance_map = self._precondition_instances
        for precondition, _ in self._method.precondition:
            self.traverse(precondition)
        self._current_instance_map = None

    def _traverse_postconditions(self) -> None:
        """Collect all needed information about obligations."""
        assert self._current_instance_map is None
        self._current_instance_map = self._postcondition_instances
        for postcondition, _ in self._method.postcondition:
            self.traverse(postcondition)
        self._current_instance_map = None

    def _traverse_declared_exceptions(self) -> None:
        """Collect all needed information about obligations."""
        assert self._current_instance_map is None
        self._current_instance_map = self._postcondition_instances
        for postconditions in self._method.declared_exceptions.values():
            for postcondition, _ in postconditions:
                self.traverse(postcondition)
        self._current_instance_map = None

    def _get_must_terminate_instances(self) -> List[GuardedObligationInstance]:
        return self.get_precondition_instances(
            self._obligation_manager.must_terminate_obligation.identifier())

    def _check_must_terminate_measure_decrease(
            self, measure: sil.IntExpression) -> sil.BoolExpression:
        return self.caller_measure_map.check(
            sil.RefVar(self.current_thread_var), measure)

    def get_precondition_instances(
            self, obligation_id: str) -> List[ObligationInstance]:
        """Return precondition instances of specific obligation type."""
        return self._precondition_instances[obligation_id]

    def get_all_precondition_instances(self) -> List[ObligationInstance]:
        """Return all precondition instances."""
        all_instances = []
        for instances in self._precondition_instances.values():
            all_instances.extend(instances)
        return all_instances

    def add_precondition(self, expr: Expr) -> None:
        """Add additional precondition at the end."""
        self._additional_preconditions.append(expr)

    def get_additional_preconditions(self) -> List[Expr]:
        """Return additional preconditions."""
        return self._additional_preconditions

    def add_postcondition(self, expr: Expr) -> None:
        """Add additional postcondition at the end."""
        self._additional_postconditions.append(expr)

    def get_additional_postconditions(self) -> List[Expr]:
        """Return additional postconditions."""
        return self._additional_postconditions


class PythonLoopObligationInfo(BaseObligationInfo):
    """Info about the obligation use in a loop."""

    def __init__(
            self, obligaton_manager: ObligationManager,
            node: ast.While, translator: 'AbstractTranslator',
            method: PythonMethod, err_var: PythonVar = None) -> None:
        super().__init__(obligaton_manager, method)
        self.node = node
        self._instances = dict(
            (obligation.identifier(), [])
            for obligation in self._obligation_manager.obligations)
        loop_measure_map_var = self._create_measure_var(
            MEASURES_LOOP_NAME, translator,
            local=not obligation_config.disable_measures)
        self.loop_measure_map = MeasureMap(loop_measure_map_var)
        self._variables = {
            LOOP_CHECK_BEFORE_NAME: self._create_var(
                LOOP_CHECK_BEFORE_NAME, PRIMITIVE_BOOL_TYPE,
                translator.translator, local=True),
            LOOP_TERMINATION_FLAG_NAME: self._create_var(
                LOOP_TERMINATION_FLAG_NAME, PRIMITIVE_BOOL_TYPE,
                translator.translator, local=True),
            LOOP_ORIGINAL_MUST_TERMINATE_AMOUNT_NAME: self._create_perm_var(
                LOOP_ORIGINAL_MUST_TERMINATE_AMOUNT_NAME, translator,
                local=True),
            RESIDUE_LEVEL_LOOP_NAME: self._create_perm_var(
                RESIDUE_LEVEL_LOOP_NAME, translator, local=True),
        }
        self.iteration_err_var = err_var
        """In the for loop translation holds ``__iter__`` result."""
        self._additional_invariants = []
        self._prepend_body = []
        self._after_loop = []

    @property
    def loop_check_before_var(self) -> PythonVar:
        """Return variable that indicates exhaling before loop."""
        return self._variables[LOOP_CHECK_BEFORE_NAME]

    @property
    def termination_flag_var(self) -> PythonVar:
        """Return variable that indicates if loop promised to terminate."""
        return self._variables[LOOP_TERMINATION_FLAG_NAME]

    @property
    def original_must_terminate_var(self) -> PythonVar:
        """Return variable that holds original termination amount."""
        return self._variables[LOOP_ORIGINAL_MUST_TERMINATE_AMOUNT_NAME]

    @property
    def residue_level(self) -> PythonVar:
        """Return variable that represents the residue level."""
        return self._variables[RESIDUE_LEVEL_LOOP_NAME]

    @property
    def current_thread_var(self) -> PythonVar:
        """Return the variable that denotes current thread in method."""
        return self._method.obligation_info.current_thread_var

    def traverse_invariants(self) -> None:
        """Collect all needed information about obligations."""
        assert self._current_instance_map is None
        self._current_instance_map = self._instances
        for invariant, _ in self._method.loop_invariants[self.node]:
            if isinstance(invariant, ast.Expr):
                self.traverse(invariant.value.args[0])
            else:
                self.traverse(invariant.args[0])
        self._current_instance_map = None

    def _get_must_terminate_instances(self) -> List[GuardedObligationInstance]:
        return self.get_instances(
            self._obligation_manager.must_terminate_obligation.identifier())

    def _check_must_terminate_measure_decrease(
            self, measure: sil.IntExpression) -> sil.BoolExpression:
        return self.loop_measure_map.check(
            sil.RefVar(self.current_thread_var), measure)

    def get_all_instances(self) -> List[ObligationInstance]:
        """Return all invariant instances."""
        return list(self._all_instances.values())

    def get_instances(self, obligation_id: str) -> List[ObligationInstance]:
        """Return invariant instances of specific obligation type."""
        return self._instances[obligation_id]

    def construct_loop_condition(self) -> sil.BoolExpression:
        """Construct loop condition."""
        if isinstance(self.node, ast.While):
            return sil.PythonBoolExpression(self.node.test)
        else:
            return sil.RefVar(self.iteration_err_var) == None  # noqa: E711

    def prepend_body(self, stmt: Stmt) -> None:
        """Add statement at the beginning of loop body."""
        self._prepend_body.append(stmt)

    def get_prepend_body(self) -> List[Stmt]:
        """Return statements to be prepended to the loop body."""
        return self._prepend_body

    def append_after_loop(self, stmt: Stmt) -> None:
        """Add statement after the loop."""
        self._after_loop.append(stmt)

    def get_after_loop(self) -> List[Stmt]:
        """Get statements to be appended after the loop."""
        return self._after_loop

    def add_invariant(self, expr: Expr) -> None:
        """Add the additional invariant at the end."""
        self._additional_invariants.append(expr)

    def get_additional_invariants(self) -> List[Expr]:
        """Return additional invariants."""
        return self._additional_invariants



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import logging
import mypy.build
import os

from mypy.build import BuildSource
from nagini_translation.lib import config
from nagini_translation.lib.constants import IGNORED_IMPORTS, LITERALS
from nagini_translation.lib.util import (
    construct_lambda_prefix,
)
from typing import List, Optional


logger = logging.getLogger('nagini_translation.lib.typeinfo')


def col(node) -> Optional[int]:
    """
    Returns the column in a mypy mypy AST node, if any.
    """
    if hasattr(node, 'column'):
        return node.column
    return None


class TypeException(Exception):
    def __init__(self, messages):
        self.messages = messages


class TypeVisitor(mypy.traverser.TraverserVisitor):
    def __init__(self, type_map, path, ignored_lines):
        self.prefix = []
        self.all_types = {}
        self.alt_types = {}
        self.type_map = type_map
        self.path = path
        self.ignored_lines = ignored_lines
        self.type_aliases = {}
        self.type_vars = {}

    def _is_result_call(self, node: mypy.nodes.Node) -> bool:
        """Checks if call is either ``Result`` or ``RaisedException``."""
        if isinstance(node, mypy.nodes.CallExpr):
            if node.callee.name == 'Result':
                return True
            if node.callee.name == 'RaisedException':
                return True
        return False

    def visit_decorator(self, node: mypy.nodes.Decorator):
        self.visit_func_def(node.func)

    def visit_member_expr(self, node: mypy.nodes.MemberExpr):
        rectype = self.type_of(node.expr)
        if (not self._is_result_call(node.expr) and
                not isinstance(node.expr, mypy.nodes.IndexExpr) and
                not isinstance(rectype, mypy.types.CallableType) and
                not isinstance(rectype, str) and
                not isinstance(rectype, mypy.types.AnyType) and
                not isinstance(rectype, mypy.types.TypeVarType)):
            if isinstance(rectype, mypy.types.UnionType):
                # Collext all non-None elements of the union,
                # taking nested unions into account.
                utypes = [rectype]
                uindex = 0
                types = []
                while uindex < len(utypes):
                    for i in utypes[uindex].items:
                        if isinstance(i, mypy.types.UnionType):
                            utypes.append(i)
                        elif not isinstance(i, mypy.types.NoneTyp):
                            types.append(i)
                    uindex += 1
            else:
                types = [rectype]
            for t in types:
                if not hasattr(t, 'type'): # Work around issue 979 in MyPy
                    t = t.fallback   # 'TupleType' object has no attribute 'type'
                self.set_type(t.type.fullname().split('.') + [node.name],
                              self.type_of(node),
                              node.line, col(node))
        super().visit_member_expr(node)

    def visit_del_stmt(self, node: mypy.nodes.DelStmt):
        pass

    def visit_try_stmt(self, node: mypy.nodes.TryStmt):
        for var in node.vars:
            if var is not None:
                self.set_type(self.prefix + [var.name], self.type_of(var),
                              var.line, col(var))
        super().visit_try_stmt(node)

    def visit_assignment_stmt(self, node: mypy.nodes.AssignmentStmt):
        if (isinstance(node.rvalue, mypy.nodes.IndexExpr) and
                isinstance(node.rvalue.analyzed, mypy.nodes.TypeAliasExpr)):
            # If it's a type alias, process it as such.
            key = tuple(self.prefix + [node.lvalues[0].name])
            self.type_aliases[key] = node.rvalue.analyzed.type
        elif (isinstance(node.rvalue, mypy.nodes.CallExpr) and
                isinstance(node.rvalue.analyzed, mypy.nodes.TypeVarExpr)):
            key = tuple(self.prefix + [node.rvalue.analyzed._name])
            self.type_vars[key] = (node.rvalue.analyzed.upper_bound,
                                   node.rvalue.analyzed.values)
        else:
            super().visit_assignment_stmt(node)

    def visit_name_expr(self, node: mypy.nodes.NameExpr):
        is_alias = False
        for i in range(len(self.prefix)):
            key = tuple(self.prefix[:i] + [node.name])
            if key in self.type_aliases:
                is_alias = True
                break
        if (node.name not in LITERALS and not is_alias):
            name_type = self.type_of(node)
            if not isinstance(name_type, mypy.types.CallableType):
                self.set_type(self.prefix + [node.name], name_type,
                              node.line, col(node))

    def visit_star_expr(self, node: mypy.nodes.StarExpr):
        node.expr.accept(self)

    def visit_func_def(self, node: mypy.nodes.FuncDef):
        oldprefix = self.prefix
        self.prefix = self.prefix + [node.name()]
        functype = self.type_of(node)
        self.set_type(self.prefix, functype, node.line, col(node), True)
        for arg in node.arguments:
            self.set_type(self.prefix + [arg.variable.name()],
                          arg.variable.type, arg.line, col(arg))
        super().visit_func_def(node)
        self.prefix = oldprefix

    def visit_func_expr(self, node: mypy.nodes.FuncExpr):
        oldprefix = self.prefix
        prefix_string = construct_lambda_prefix(node.line, col(node))
        self.prefix = self.prefix + [prefix_string]
        for arg in node.arguments:
            self.set_type(self.prefix + [arg.variable.name()],
                          arg.variable.type, arg.line, col(arg))
        node.body.accept(self)
        self.prefix = oldprefix

    def visit_class_def(self, node: mypy.nodes.ClassDef):
        oldprefix = self.prefix
        self.prefix = self.prefix + [node.name]
        super().visit_class_def(node)
        self.prefix = oldprefix

    def set_type(self, fqn, type, line, col, return_type=False):
        if return_type and isinstance(type, mypy.types.CallableType):
            type = type.ret_type
        if not type or isinstance(type, mypy.types.AnyType):
            if line in self.ignored_lines:
                return
            else:
                error = ' error: Encountered Any type. Type annotation missing?'
                msg = ':'.join([self.path, str(line), error])
                raise TypeException([msg])
        key = tuple(fqn)
        if key in self.all_types:
            if not self.type_equals(self.all_types[key], type):
                # Type change after isinstance
                if key not in self.alt_types:
                    self.alt_types[key] = {}
                self.alt_types[key][(line, col)] = type
                return
        self.all_types[key] = type

    def type_equals(self, t1, t2):
        if str(t1) == str(t2):
            return True
        if (isinstance(t1, mypy.types.FunctionLike) and
                isinstance(t2, mypy.types.FunctionLike)):
            if self.type_equals(t1.ret_type, t2.ret_type):
                all_eq = True
                for arg1, arg2 in zip(t1.arg_types, t2.arg_types):
                    all_eq = all_eq and self.type_equals(arg1, arg2)
                return all_eq
        return t1 == t2

    def visit_call_expr(self, node: mypy.nodes.CallExpr):
        if (isinstance(node.callee, mypy.nodes.NameExpr) and
                    node.callee.fullname == 'typing.cast'):
            return
        for a in node.args:
            a.accept(self)
        node.callee.accept(self)

    def type_of(self, node):
        if hasattr(node, 'node') and isinstance(node.node, mypy.nodes.MypyFile):
            return node.fullname
        if isinstance(node, mypy.nodes.FuncDef):
            if node.type:
                return node.type
        if isinstance(node, mypy.nodes.NameExpr):
            key = (node.name,)
            if key in self.all_types:
                return self.all_types[key]
        elif isinstance(node, mypy.nodes.CallExpr):
            if node.callee.name == 'Result':
                key = tuple(self.prefix)
                for i in range(len(key)):
                    if key[i].startswith('lambda'):
                        key = key[:i]
                        break
                type = self.all_types[key]
                return type
        if node in self.type_map:
            result = self.type_map[node]
            return result
        else:
            msg = self.path + ':' + str(node.get_line()) + ': error: '
            if isinstance(node, mypy.nodes.FuncDef):
                msg += 'Encountered Any type. Type annotation missing?'
            else:
                msg += 'dead.code'
            raise TypeException([msg])

    def visit_comparison_expr(self, o: mypy.nodes.ComparisonExpr):
        # Weird things seem to happen with is-comparisons, so we ignore those.
        if 'is' not in o.operators and 'is not' not in o.operators:
            super().visit_comparison_expr(o)


class TypeInfo:
    """
    Provides type information for all variables and functions in a given
    Python module.
    """

    def __init__(self):
        self.all_types = {}
        self.alt_types = {}
        self.files = {}
        self.type_aliases = {}
        self.type_vars = {}

    def _create_options(self, strict_optional: bool):
        """
        Creates an Options object for mypy and activates strict optional typing
        based on the given argument.
        As long as mypy actually ignores these options, this will also set
        the STRICT_OPTIONAL flag in the experimental module to the given value.
        """
        result = mypy.options.Options()
        result.strict_optional = strict_optional
        result.show_none_errors = strict_optional
        result.show_traceback = True
        # This is an experimental feature atm and you actually have to
        # enable it like this
        mypy.experiments.STRICT_OPTIONAL = strict_optional
        result.fast_parser = True
        return result

    def check(self, filename: str) -> bool:
        """
        Typechecks the given file and collects all type information needed for
        the translation to Viper
        """

        def report_errors(errors: List[str]) -> None:
            for error in errors:
                logger.info(error)
            raise TypeException(errors)

        try:
            options_strict = self._create_options(True)
            res_strict = mypy.build.build(
                [BuildSource(filename, None, None)],
                options_strict, bin_dir=config.mypy_dir
                )

            if res_strict.errors:
                # Run mypy a second time with strict optional checking disabled,
                # s.t. we don't get overapproximated none-related errors.
                options_non_strict = self._create_options(False)
                res_non_strict = mypy.build.build(
                    [BuildSource(filename, None, None)],
                    options_non_strict, bin_dir=config.mypy_dir
                )
                if res_non_strict.errors:
                    report_errors(res_non_strict.errors)
            for name, file in res_strict.files.items():
                if name in IGNORED_IMPORTS:
                    continue
                self.files[name] = file.path
                visitor = TypeVisitor(res_strict.types, name,
                                      file.ignored_lines)
                visitor.prefix = name.split('.')
                file.accept(visitor)
                self.all_types.update(visitor.all_types)
                self.alt_types.update(visitor.alt_types)
                self.type_aliases.update(visitor.type_aliases)
                self.type_vars.update(visitor.type_vars)
            return True
        except mypy.errors.CompileError as e:
            report_errors(e.messages)

    def get_type_prefix(self, name: str) -> str:
        name = os.path.abspath(name)
        for prefix, path in self.files.items():
            path = os.path.abspath(path)
            if path == name:
                return prefix
        return None

    def get_type(self, prefix: List[str], name: str):
        """
        Looks up the inferred or annotated type for the given name in the given
        prefix
        """
        key = tuple(prefix + [name])
        result = self.all_types.get(key)
        alts = self.alt_types.get(key)
        if result is None:
            if not prefix:
                return None, None
            else:
                return self.get_type(prefix[:len(prefix) - 1], name)
        else:
            return result, alts

    def get_func_type(self, prefix: List[str]):
        """
        Looks up the type of the function which creates the given context
        """
        result = self.all_types.get(tuple(prefix))
        if result is None:
            if len(prefix) == 0:
                return None
            else:
                return self.get_func_type(prefix[:len(prefix) - 1])
        else:
            if isinstance(result, mypy.types.FunctionLike):
                result = result.ret_type
            return result

    def is_normal_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.nodes.TypeInfo)

    def is_instance_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.Instance)

    def is_tuple_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.TupleType)

    def is_void_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.Void)

    def is_union_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.UnionType)

    def is_callable_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.CallableType)

    def is_type_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.TypeType)

    def is_type_var(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.TypeVarType)

    def is_none_type(self, type: mypy.types.Type) -> bool:
        return isinstance(type, mypy.types.NoneTyp)
